"""
// Copyright (C) 2022-2024 MLACS group (AC)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

import numpy as np

from ase.io import Trajectory
from ase.units import fs
from ase.md.langevin import Langevin
from ase.md import MDLogger
from ase.md.velocitydistribution import MaxwellBoltzmannDistribution
from ase.calculators.lammpsrun import LAMMPS

from ..core.manager import Manager
from .state import StateManager


# ========================================================================== #
# ========================================================================== #
class LangevinState(StateManager):
    """
    State Class for running a Langevin simulation as implemented in ASE

    Parameters
    ----------
    temperature : :class:`float`
        Temperature of the simulation, in Kelvin

    friction : :class:`float` (optional)
        Friction coefficient of the thermostat. Default 0.01.

    dt : :class:`float` (optional)
        Timestep, in fs. Default ``1.5`` fs.

    nsteps : :class:`int` (optional)
        Number of MLMD steps for production runs. Default ``1000`` steps.

    nsteps_eq : :class:`int` (optional)
        Number of MLMD steps for equilibration runs. Default ``100`` steps.

    fixcm : :class:`Bool` (optional)
        Fix position and momentum center of mass. Default True.

    logfile : :class:`str` (optional)
        Name of the file for logging the MLMD trajectory.
        If ``None``, no log file is created. Default ``None``.

    trajfile : :class:`str` (optional)
        Name of the file for saving the MLMD trajectory.
        If ``None``, no traj file is created. Default ``None``.

    loginterval : :class:`int` (optional)
        Number of steps between MLMD logging. Default ``50``.

    rng : RNG object (optional)
        Rng object to be used with the Langevin thermostat.
        Default correspond to ``numpy.random.default_rng()``.

    init_momenta : :class:`numpy.ndarray` (optional)
        If ``None``, velocities are initialized with a
        Maxwell Boltzmann distribution
        N * 3 velocities for the initial configuration
    """
    def __init__(self,
                 temperature,
                 friction=0.01,
                 dt=1.5,
                 nsteps=1000,
                 nsteps_eq=100,
                 fixcm=True,
                 logfile=None,
                 trajfile=None,
                 loginterval=50,
                 rng=None,
                 init_momenta=None,
                 **kwargs):

        StateManager.__init__(self,
                              nsteps=nsteps,
                              nsteps_eq=nsteps_eq,
                              logfile=logfile,
                              trajfile=trajfile,
                              loginterval=loginterval,
                              **kwargs)
        self.pressure = None
        self.dt = dt
        self.fixcm = fixcm
        self.temperature = temperature
        self.dt = dt
        self.fixcm = fixcm
        self.friction = friction
        self.init_momenta = init_momenta
        self.rng = rng
        if self.rng is None:
            self.rng = np.random.default_rng()
        self.pressure = None

        self.info_dynamics = dict(ensemble="NVT",
                                  temperature=self.temperature)

        self.ispimd = False
        self.isrestart = False
        self.isneb = False

# ========================================================================== #
    @Manager.exec_from_subsubdir
    def run_dynamics(self,
                     supercell,
                     pair_style,
                     pair_coeff,
                     model_post=None,
                     atom_style="atomic",
                     eq=False,
                     nbeads=1,
                     elements=None):
        """
        """
        atoms = supercell.copy()
        calc = LAMMPS(pair_style=pair_style, pair_coeff=pair_coeff,
                      atom_style=atom_style,
                      # keep_tmp_files=True, # DEBUG
                      # tmp_dir=str(self.workdir/'tmp_lammps'),  # DEBUG
                      )

        if model_post is not None:
            calc.set(model_post=model_post)

        atoms.calc = calc

        if eq:
            nsteps = self.nsteps_eq
        else:
            nsteps = self.nsteps

        dyn = Langevin(atoms,
                       self.dt*fs,
                       temperature_K=self.temperature,
                       friction=self.friction,
                       fixcm=self.fixcm,
                       rng=self.rng)
        if self.trajfile is not None:
            trajectory = Trajectory(self.trajfile, mode="a", atoms=atoms)
            dyn.attach(trajectory.write, interval=self.loginterval)

        if self.logfile is not None:
            dyn.attach(MDLogger(dyn, atoms, self.logfile, stress=True),
                       interval=self.loginterval)
        dyn.run(nsteps)
        dyn.atoms.info['info_state'] = self.info_dynamics
        return dyn.atoms

# ========================================================================== #
    def initialize_momenta(self, atoms):
        """
        """
        if self.init_momenta is None:
            MaxwellBoltzmannDistribution(atoms,
                                         temperature_K=self.temperature,
                                         rng=self.rng)
        else:
            atoms.set_momenta(self.init_momenta)

# ========================================================================== #
    def log_recap_state(self):
        """
        """
        msg = "NVT Langevin dynamics as implemented in ASE\n"
        msg += f"Temperature (in Kelvin)                {self.temperature}\n"
        msg += f"Number of MLMD equilibration steps :   {self.nsteps_eq}\n"
        msg += f"Number of MLMD production steps :      {self.nsteps}\n"
        msg += f"Timestep (in fs) :                     {self.dt}\n"
        msg += f"Friction (in fs) :                     {self.friction / fs}\n"
        msg += "\n"
        return msg



"""
// Copyright (C) 2022-2024 MLACS group (AC)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

import pytest

from ase.build import bulk

from ... import context  # noqa
from mlacs.ti import PressureScalingState


@pytest.fixture
def expected_folder(expected_folder_base):
    # return ["ThermoInt"]
    return ["ReversibleScaling"]


@pytest.fixture
def expected_files(expected_files_base):
    return []


def test_pressure_scaling(root, treelink):

    atoms = bulk("Cu", cubic=True).repeat(2)
    pair_style = "lj/cut 5.0"
    pair_coeff = ["* * 0.52031 2.29726"]
    temperature = 300
    folder = "ReversibleScaling"

    nsteps = 5

    state = PressureScalingState(atoms,
                                 pair_style,
                                 pair_coeff,
                                 temperature,
                                 phase="solid",
                                 nsteps=nsteps,
                                 nsteps_eq=nsteps,
                                 nsteps_msd=nsteps,
                                 nsteps_averagin=nsteps,
                                 folder=folder)
    state.run()



"""
// Copyright (C) 2022-2024 MLACS group (AC, RB, CD)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

import os
import sys
from pathlib import Path
from subprocess import check_output

import numpy as np

from ase.atoms import Atoms

from mlacs.utilities.miscellanous import create_ASE_object
from mlacs.utilities.units import unit_converter

try:
    from netCDF4 import Dataset
except ImportError as exc:
    raise ImportError("You need netCDF4 to use the AbinitNC class") from exc


# ========================================================================== #
# ========================================================================== #
class MlacsHist:
    """
    Parent class to handle the HIST.nc file created by MLACS.

    Parameters
    ----------

    ncprefix: :class:`str` (optional)
        The prefix to prepend the name of the *HIST.nc file.
        Default `''`.

    workdir: :class:`str` (optional)
        The directory in which to run the calculation.

    ncformat: :class:`str` (optional)
        The format of the *HIST.nc file. One of the five flavors of netCDF
        files format available in netCDF4 python package: 'NETCDF3_CLASSIC',
        'NETCDF3_64BIT_OFFSET', 'NETCDF3_64BIT_DATA','NETCDF4_CLASSIC',
        'NETCDF4'.
        Default ``NETCDF3_CLASSIC``.

    launched: :class:`Bool` (optional)
        If True then is not the first MLACS start of the related Mlas instance,
        i.e. it is a restart situation for which a *HIST.nc already exists.
        Default ``True``.

    atoms: :class:`ase.Atoms` or :class:`list` of :class:`ase.Atoms` (optional)
        the atom object on which the simulation is run.
        Default ``None``.

    ncpath: :class:`str` or :class:`Path` of `pathlib` module (optional)
        Absolute path to HIST.nc file, i.e. `path_to_ncfile/ncfilename`.
        Compulsory in case of post-processing usage.
        Default ``None``.

    Examples
    -------
    Post-processing usage:
    ::

        ncfile = MlacsHist(ncpath='/path/to/HIST.nc')
        var_names = ncfile.get_var_names()
        dict_var_units = ncfile.get_units()
        var_dim_dict = ncfile.var_dim_dict
        energy_array = ncfile.read_obs('etotal')

    Conversion of a list of ASE Atoms into HIST.nc:
    ::

        ncprefix = 'my_ncprefix'
        workdir = '/path/to/workdir/'
        obj = MlacsHist(ncprefix=ncprefix,
                        workdir=workdir,
                        atoms=list_of_ase_atoms)
        obj.convert_to_hist()

    The HIST.nc is saved in workdir with name f'{ncprefix}_HIST.nc'.
    """

    def __init__(self,
                 ncprefix='',
                 workdir='',
                 ncformat='NETCDF3_CLASSIC',
                 launched=True,
                 atoms=None,
                 ncpath=None):

        self.ncprefix = ncprefix
        self.workdir = workdir
        self.ncformat = ncformat
        self.launched = launched
        self.atoms = atoms
        self._set_name_conventions()
        self._set_unit_conventions()

        if ncpath:
            self.ncpath = ncpath
            self.workdir = Path(ncpath).parents[0]

            # Specific to post-processing usage
            if Path(ncpath).is_file():
                with Dataset(str(ncpath), 'r') as ncfile:
                    self.ncformat = ncfile.file_format

# ========================================================================== #
    def _initialize_nc_file(self, atoms):
        """
        Initialize netCDF file:
            - Create Abinit-style dimensions.
            - Create Abinit-style variables that are not 'CalcProperty'. The
            latter are typically static structural data obeying Abinit naming
            conventions. Some of these variables (namely: znucl, typat, amu,
            dtion) are initialized here.

        Parameters
        ----------

        atoms: :class:`ase.Atoms` or :class:`list` of :class:`ase.Atoms`
            An example of the atom object produced by the simulation. It is
            assumed the all ase.Atoms` have the same  `typat`, `znucl`, etc.
            This excludes, e.g., grand-canonical computations.
        """

        # Assume ntypat, natom, etc., are the same for all items in list
        if isinstance(atoms, list):
            atoms = atoms[0]

        atomic_numbers = list(atoms.get_atomic_numbers())
        atomic_masses = list(atoms.get_masses())
        natom = len(atoms)
        znucl = sorted(set(atomic_numbers), key=atomic_numbers.index)
        amu = sorted(set(atomic_masses), key=atomic_masses.index)
        ntypat = len(znucl)
        typat = [1+znucl.index(x) for x in atomic_numbers]
        # dtion is not well-defined in MLACS. Set to one below by convention.
        dtion = 1.0

        dict_dim = {'time': None,
                    'one': 1,
                    'two': 2,
                    'xyz': 3,
                    'npsp': 3,
                    'six': 6,
                    'ntypat': ntypat,
                    'natom': natom,
                    }
        dict_var = {'typat': ('natom',),
                    'znucl': ('ntypat',),
                    'amu': ('ntypat',),
                    'dtion': ('one',),
                    'mdtemp': ('two',),
                    'mdtime': ('time',),
                    }
        dict_initialize_var = {'typat': typat,
                               'znucl': znucl,
                               'amu': amu,
                               'dtion': dtion,
                               }

        self._add_dim(self.ncpath, dict_dim)
        self._add_var(self.ncpath, dict_var)
        self._initialize_var(self.ncpath, dict_initialize_var)

# ========================================================================== #
    def _add_dim(self, ncfilepath, dict_dim, mode='r+'):
        with Dataset(ncfilepath, mode, format=self.ncformat) as new:
            for dim_name, dim_value in dict_dim.items():
                new.createDimension(dim_name, (dim_value))

# ========================================================================== #
    def _add_var(self, ncfilepath, dict_var, mode='r+', datatype='float64'):
        with Dataset(ncfilepath, mode, format=self.ncformat) as new:
            for var_name, var_dim in dict_var.items():
                new.createVariable(var_name, datatype, var_dim)

# ========================================================================== #
    def _initialize_var(self, ncfilepath, dict_initialize_var, mode='r+'):
        with Dataset(ncfilepath, mode, format=self.ncformat) as new:
            for var_name, var_value in dict_initialize_var.items():
                new[var_name][:] = var_value

# ========================================================================== #
    def create_nc_var(self, prop_list):
        """
        Create Abinit-style variables in netCDF file from `prop_list`.

        Parameters
        -------

        prop_list: :class:`list` of :class: `CalcProperty`
            Cf. mlacs.properties.calc_property.py
        """
        datatype = 'float64'
        if prop_list is not None:
            for obs in prop_list:
                nc_name = obs.nc_name
                nc_dim = obs.nc_dim
                # Observables need these attributes to get saved in the HIST.nc
                if None not in (nc_name, nc_dim):
                    with Dataset(self.ncpath, 'a') as ncfile:
                        var = ncfile.createVariable(nc_name, datatype, nc_dim)
                        var.setncattr('unit', obs.nc_unit)
                        meta_dim = ('time', 'two',)
                        meta_name = nc_name + '_meta'
                        ncfile.createVariable(meta_name, datatype, meta_dim)
                        w_name = 'weighted_' + nc_name
                        wvar = ncfile.createVariable(w_name, datatype, nc_dim)
                        wvar.setncattr('unit', obs.nc_unit)

# ========================================================================== #
    def read_obs(self, obs_name):
        """Read specific observable from netcdf file"""
        with Dataset(self.ncpath, 'r') as ncfile:
            observable_values = ncfile[obs_name][:].data
        return observable_values

# ========================================================================== #
    def read_weighted_obs(self, obs_name):
        """
        Read specific weighted observable from netcdf file.
        Return values, idx in database
        """
        with Dataset(self.ncpath, 'r') as ncfile:
            wobs_values = ncfile[obs_name][:]
            weighted_obs_data = wobs_values[wobs_values.mask == False].data  # noqa
            weighted_obs_idx = 1 + np.where(~wobs_values.mask)[0]
        return weighted_obs_data, weighted_obs_idx

# ========================================================================== #
    def _set_name_conventions(self):
        """Define naming conventions related to routine properties"""
        # Variable names and dimensions as defined in Abinit
        var_dim_dict = {'Total_Energy': ['etotal', ('time',)],
                        'Kinetic_Energy': ['ekin', ('time',)],
                        'Potential_Energy': ['epot', ('time',)],
                        'Velocities': ['vel', ('time', 'natom', 'xyz')],
                        'Forces': ['fcart', ('time', 'natom', 'xyz')],
                        'Positions': ['xcart', ('time', 'natom', 'xyz')],
                        'Scaled_Positions': ['xred', ('time', 'natom', 'xyz')],
                        'Temperature': ['temper', ('time',)],
                        'Volume': ['vol', ('time',)],
                        'Stress': ['strten', ('time', 'six')],
                        'Cell': ['rprimd', ('time', 'xyz', 'xyz')]
                        }
        self.var_dim_dict = var_dim_dict

# ========================================================================== #
    def _set_unit_conventions(self):
        """Define unit conventions related to routine properties"""
        # XXX : This should be moved to utilities.units.py
        # Dict whose keys are 'var names' and values are ASE units
        ase_units_dict = {'Total_Energy': 'eV',
                          'Kinetic_Energy': 'eV',
                          'Potential_Energy': 'eV',
                          'Velocities': '',
                          'Forces': 'eV/Ang',
                          'Positions': 'Ang',
                          'Scaled_Positions': 'dimensionless',
                          'Temperature': 'K',
                          'Volume': 'Ang^3',
                          'Stress': 'eV/Ang^3',
                          'Cell': 'Ang',
                          }
        self.ase_units_dict = ase_units_dict

        # Dict whose keys are 'var names' and values are Abinit units
        abinit_units_dict = {'Total_Energy': 'Ha',
                             'Kinetic_Energy': 'Ha',
                             'Potential_Energy': 'Ha',
                             'Velocities': '',
                             'Forces': 'Ha/Bohr',
                             'Positions': 'Bohr',
                             'Scaled_Positions': 'dimensionless',
                             'Temperature': 'K',
                             'Volume': 'Bohr^3',
                             'Stress': 'Ha/Bohr^3',
                             'Cell': 'Bohr',
                             }
        self.abinit_units_dict = abinit_units_dict

# ========================================================================== #
    def read_all(self):
        """Read all observables from netcdf file"""
        res = {}
        with Dataset(self.ncpath, 'r') as ncfile:
            for name, variable in ncfile.variables.items():
                res[name] = variable[:]
        return res

# ========================================================================== #
    def get_var_names(self):
        """Return list of all observable names"""
        return list(self.read_all().keys())

# ========================================================================== #
    def get_units(self):
        """
        Return dict where keys are obs. names and values are units.
        Variables without units do not appear.
        """
        res = {}
        with Dataset(self.ncpath, 'r') as ncfile:
            for name, variable in ncfile.variables.items():
                if hasattr(variable, 'unit'):
                    res[name] = variable.unit
        return res

# ========================================================================== #
    def convert_to_hist(self):
        """
        Convert list of Ase's Atoms into HIST.nc respecting Abinit conventions.
        """

        # Ensure `self.atoms` is a list
        if isinstance(self.atoms, Atoms):
            atoms_list = [self.atoms]
        atoms_list = self.atoms

        # Initialize target HIST.nc file
        ncname = self.ncprefix + "_HIST.nc"
        ncpath = Path(self.workdir) / ncname
        self.ncpath = ncpath
        self._initialize_nc_file(atoms_list[0])

        # Initialize list of properties
        # TODO: Fix circular dependency of RoutinePropertyManager import
        from mlacs.properties import RoutinePropertyManager
        rout_prop_obj = RoutinePropertyManager(self, False)

        # Emulate Mlas run
        for idx, at in enumerate(atoms_list):
            rout_prop_obj.calc_initialize(atoms=[at])
            step = idx + 1
            rout_prop_obj.run(step)
            rout_prop_obj.save_prop(step)


# ========================================================================== #
# ========================================================================== #
class OtfMlacsHist(MlacsHist):
    """
    Handle HIST.nc file during the on-the-fly execution of MLACS.

    This class is thought as a low-level, internal object used only during the
    MLACS cycle, through the `mlas` object. It is tailored to save 'Properties'
    'RoutineProperties', weights and metadata, in a common HIST.nc file.

    Notes
    ----------

    The Hist name format is: `ncprefix + scriptname + '_HIST.nc'.`

    Parameters
    ----------

    ncprefix: :class:`str` (optional)
        The prefix to prepend the name of the *HIST.nc file.
        Default `''`.

    workdir: :class:`str` (optional)
        The directory in which to run the calculation.

    ncformat: :class:`str` (optional)
        The format of the *HIST.nc file. One of the five flavors of netCDF
        files format available in netCDF4 python package: 'NETCDF3_CLASSIC',
        'NETCDF3_64BIT_OFFSET', 'NETCDF3_64BIT_DATA','NETCDF4_CLASSIC',
        'NETCDF4'.
        Default ``NETCDF3_CLASSIC``.

    launched: :class:`Bool` (optional)
        If True then is not the first MLACS start of the related Mlas instance,
        i.e. it is a restart situation for which a *HIST.nc already exists.
        Default ``True``.

    atoms: :class:`ase.Atoms` or :class:`list` of :class:`ase.Atoms` (optional)
        the atom object on which the simulation is run.
        Default ``None``.
    """

    def __init__(self,
                 ncprefix='',
                 workdir='',
                 ncformat='NETCDF3_CLASSIC',
                 launched=True,
                 atoms=None):
        MlacsHist.__init__(self,
                           ncprefix=ncprefix,
                           workdir=workdir,
                           ncformat=ncformat,
                           launched=launched,
                           atoms=atoms,
                           ncpath=None)
        self.ncformat = ncformat
        self.workdir = workdir
        self.ncpath = self._get_nc_path()
        # Check if there is only one type of ase.Atoms objects
        all_chem_symb = [at.get_chemical_formula() for at in atoms]
        self.unique_atoms_type = len(set(all_chem_symb)) == 1
        if not os.path.isfile(self.ncpath) and self.unique_atoms_type:
            self._initialize_nc_file(atoms)
            self._create_weights_var_dim()

# ========================================================================== #
    def _get_nc_path(self):
        """Return netcdf path of Abinit-style HIST file."""
        ncpref = self.ncprefix
        wdir = self.workdir
        if ncpref and not ncpref.endswith('_'):
            ncpref += '_'

        script_name = ncpref
        pytest_path = os.getenv('PYTEST_CURRENT_TEST')
        if pytest_path:
            wdir = wdir or Path(pytest_path).parents[0].absolute()
            script_name += Path(pytest_path).stem
        else:
            script_name += Path(sys.argv[0]).stem
        ncname = script_name + "_HIST.nc"

        # Deal with potential duplicates, i.e. existing files with ncname
        ncpath = Path(wdir).absolute() / ncname
        if ncpath.is_file():
            if not self.launched:   # if first MLAS launch
                suffix = 1
                while ncpath.is_file():
                    ncname = f"{script_name}_{suffix:04d}_HIST.nc"
                    ncpath = Path(wdir) / ncname
                    suffix += 1

        return str(ncpath)

# ========================================================================== #
    def _create_weights_var_dim(self):
        dict_w_dim = {'weights_dim': None}
        dict_w_var = {'weights': ('weights_dim',),
                      'weights_meta': ('weights_dim', 'xyz',),
                      }

        self.weights_ncpath = self.ncpath
        if 'NETCDF3' in self.ncformat:
            dict_w_dim['xyz'] = 3
            self.weights_ncpath = self.ncpath.replace('HIST', 'WEIGHTS')

        self._add_dim(self.weights_ncpath, dict_w_dim)
        self._add_var(self.weights_ncpath, dict_w_var)


# ========================================================================== #
# ========================================================================== #
class AbinitNC:
    """
    Class to read all netCDF files created by Abinit.

    Can also convert a netCDF file into a list of ASE Atoms objects, cf.
    convert_to_atoms().

    Parameters
    ----------

    workdir: :class:`str` (optional)
        The root for the directory.
        Default 'DFT'

    prefix: :class:`str` (optional)
        The prefix of Abinit nc file, i.e. str before o_{suffix}.nc.

    suffix: :class:`str` (optional)
        'HIST' or 'GSR' or 'OUT'.
        If no suffix is given, it is set automatically by _update_suffix().

    Examples
    -------
    netCDF files can be loaded by defining the `ncfile` attribute, here for
    instance with a HIST.nc:
    ::

       hist = AbinitNC()
       hist.ncfile = '/path/to/HIST.nc'
       atoms_list = hist.convert_to_atoms()

    Alternatively, one can directly pass workdir, prefix, suffix as parameters
    of the AbinitNC class. The path of the ncfile must follow the syntax
    f'{workdir}{prefix}o_{suffix}.nc'. For instance with a GSR.nc:
    ::

        gsr = AbinitNC(workdir=workdir, prefix=prefix, suffix='GSR')
        atoms_list = gsr.convert_to_atoms()

    Once the netCDF files are loaded, the data can be extracted as a dictionary
    where keys are Abinit variable names and values are arrays in atomic units:
    ::

        hist_results = hist.read()
        energy = hist_results['etotal']  # Atomic units

    Alternatively, conversion to a list of ASE atoms can also be achieved in
    the following way:
    ::

        hist_atoms_list = hist.convert_to_atoms()  # ASE units
        gsr_atoms_list = gsr.convert_to_atoms()  # ASE units
    """

    def __init__(self, workdir=None, prefix='abinit', suffix=None):

        self.workdir = workdir
        if self.workdir is None:
            self.workdir = os.getcwd() + "/DFT/"
        if self.workdir[-1] != "/":
            self.workdir += "/"
        if not os.path.exists(self.workdir):
            self.workdir = ''

        self.ncfile = f'{self.workdir}{prefix}o_{suffix}.nc'
        self.results = {}
        self.atoms_list = []
        self.suffix = suffix

# ========================================================================== #
    def read(self, filename=None):
        """
        Read NetCDF output of Abinit from netCDF4 library.

        Parameters
        ----------

        filename: :class:`str` (optional)
            The name of netCDF file.

        Returns
        -------

        self.results: :class:`dict`
            The dictionary mapping netCDF variables names to their values.
        """

        if filename is not None:
            self.ncfile = filename
            self.dataset = Dataset(filename)
        elif filename is None and hasattr(self, 'ncfile'):
            self.dataset = Dataset(self.ncfile)
        else:
            raise FileNotFoundError('No NetCDF file defined')

        self._keyvar = list(self.dataset.variables)
        self._defattr()
        if not hasattr(self, 'results'):
            self.results = {}
        self.results.update(vars(self))
        return self.results

# ========================================================================== #
    def ncdump(self, filename=None) -> str:
        """
        Read netCDF output of Abinit from command-line ncdump tool.

        Parameters
        ----------

        filename: :class:`str` (optional)
            The name of netCDF file.

        Returns
        -------

        :class:`str` (optional)
            The output of the `ncdump` command.
        """
        return check_output(['ncdump', filename])

# ========================================================================== #
    def _defattr(self):
        for attr in self._keyvar:
            setattr(self, attr, self._extractattr(attr))
        for attr in self._keyvar:
            value = getattr(self, attr)
            if isinstance(value, (int, float, str)):
                continue
            if isinstance(value, np.ndarray):
                setattr(self, attr, self._decodearray(value))
            elif isinstance(value, memoryview):
                if () == value.shape:
                    setattr(self, attr, value.tolist())
                else:
                    setattr(self, attr, self._decodearray(value.obj))
            else:
                delattr(self, attr)
                msg = f'Unknown object type: {type(value)}\n'
                msg += '-> deleted attribute from AbiAtoms object.\n'
                msg += 'Should be added in the class AbiAtoms, if needed !'
                raise Warning(msg)

# ========================================================================== #
    def _extractattr(self, value):
        return self.dataset[value][:].data

# ========================================================================== #
    def _check_end_of_dataset(self, _str, consecutive_empty_spaces):
        """
        Return True if end of dataset has been reached.
        In particular, this function handles the dataset named 'input_string',
        which corresponds to the Abinit input file, but also contains unwanted
        (i.e., not encoded in UTF-8) information at the bottom.
        """
        last_lammps_line = 'chkexit 1 # abinit.exit file in the running directory'  # noqa
        last_lammps_line += ' terminates after the current SCF'
        if last_lammps_line in _str[-len(last_lammps_line):]:
            return True
        large_blank = ' '*80
        if large_blank == _str[-len(large_blank):]:
            return True
        if consecutive_empty_spaces == 80:
            return True
        return False

# ========================================================================== #
    def _decodeattr(self, value) -> str:
        _str = ''
        consec_empty_spaces = 0
        for s in value.tolist():
            if self._check_end_of_dataset(_str, consec_empty_spaces) is True:
                break
            try:
                _str += bytes.decode(s)
                if len(bytes.decode(s)) == 0:
                    consec_empty_spaces += 1
                else:
                    consec_empty_spaces = 0
            except UnicodeDecodeError:
                # Just to be on the safe side.
                break
        return _str.strip()

# ========================================================================== #
    def _decodearray(self, value):
        if 'S1' != value.dtype:
            return value
        if 1 == len(value.shape):
            return self._decodeattr(value)
        return np.r_[[self._decodeattr(v) for v in value]]

# ========================================================================== #
    def _update_suffix(self):
        """Initialize suffix when it has not been passed explicitely."""
        if 'HIST.nc' in self.ncfile:
            self.suffix = 'HIST'
        elif 'OUT.nc' in self.ncfile:
            self.suffix = 'OUT'
        elif 'GSR.nc' in self.ncfile:
            self.suffix = 'GSR'

# ========================================================================== #
    def _extract_out_data(self, results):
        """
        Extract data from Abinit's OUT.nc.
        """
        typat = results['typat'].astype(int)
        znucl = results['znucl'].astype(int)
        atomic_numbers = znucl[typat-1]
        nat = len(typat)
        rprim = np.reshape(results['rprim'], (3, 3))
        cell = results['acell'] * rprim
        positions = np.reshape(results['xcart'], (nat, 3))
        structural_data = [atomic_numbers, positions, cell]

        energy = results['etotal']
        forces = results['fcart']
        stress = results['strten']
        main_properties = [energy, forces, stress]

        free_energy = None
        if 'spinat' in results:
            spinat = results['spinat'].reshape((nat, 3))
        else:
            spinat = None
        others = [free_energy, spinat]

        return structural_data, main_properties, others

# ========================================================================== #
    def _extract_gsr_data(self, results):
        """
        Extract data from Abinit's GSR.nc.
        """
        typat = results['atom_species'].astype(int)
        znucl = results['atomic_numbers'].astype(int)
        atomic_numbers = znucl[typat-1]
        cell = results['primitive_vectors']
        positions = results['reduced_atom_positions'] @ cell
        structural_data = [atomic_numbers, positions, cell]

        energy = results['etotal']
        forces = results['cartesian_forces']
        stress = results['cartesian_stress_tensor']
        main_properties = [energy, forces, stress]

        free_energy = results['entropy']
        if 'spinat' in results:
            spinat = results['spinat']
        else:
            spinat = None
        others = [free_energy, spinat]

        return structural_data, main_properties, others

# ========================================================================== #
    def _extract_hist_data(self, results):
        """
        Extract data from Abinit's HIST.nc.
        """
        typat = results['typat'].astype(int)
        znucl = results['znucl'].astype(int)
        atomic_numbers = znucl[typat-1]
        cell = results['rprimd']
        positions = results['xcart']
        structural_data = [atomic_numbers, positions, cell]

        energy = results['etotal']
        forces = results['fcart']
        stress = results['strten']
        main_properties = [energy, forces, stress]

        free_energy = None
        spinat = None
        others = [free_energy, spinat]

        return structural_data, main_properties, others

# ========================================================================== #
    def convert_to_atoms(self):
        """
        Convert Abinit *.nc file to a list of ASE Atoms objects.

        Notes
        ----------

        In the process, atomic units (Abinit) are converted to ASE units.

        Returns
        -------

        atoms_list: :class:`list` of :class:`ase.Atoms`
            The list of configurations in ASE format.
        """

        # Ensure Abinit *.nc file has been read
        if not self.results:
            self.read()
        res_dict = self.results

        self._update_suffix()
        if self.suffix == 'HIST':
            struct_data, main_prop, others = self._extract_hist_data(res_dict)
        elif self.suffix == 'OUT':
            struct_data, main_prop, others = self._extract_out_data(res_dict)
        elif self.suffix == 'GSR':
            struct_data, main_prop, others = self._extract_gsr_data(res_dict)

        atomic_numbers, positions, cell = struct_data
        energy, forces, stress = main_prop
        free_energy, spinat = others
        if spinat is None:
            spinat = np.zeros((len(atomic_numbers), 3))

        # Add new axis to arrays to uniformize shapes with 'HIST'
        # such that array.shape becomes (nb_confs, array.shape)
        if self.suffix in ['OUT', 'GSR']:
            positions = np.expand_dims(positions, axis=0)
            cell = np.expand_dims(cell, axis=0)
            energy = np.expand_dims(energy, axis=0)
            forces = np.expand_dims(forces, axis=0)
            stress = np.expand_dims(stress, axis=0)

        # If Abinit nimage > 1, only the result of the first image is kept
        # XXX CD: in PIMD∕NEB calculations, this will need to be adapted
        if cell[0].shape != (3, 3):
            positions = positions[:, 0, ...]
            cell = cell[:, 0, ...]
            energy = energy[:, 0, ...]
            forces = forces[:, 0, ...]
            stress = stress[:, 0, ...]

        nb_confs = len(energy)
        atoms_list = []
        for i in range(nb_confs):
            # Convert atomic units to ASE units
            conv_p = unit_converter(positions[i], 'Bohr', target='ASE')[0]
            conv_c = unit_converter(cell[i], 'Bohr', target='ASE')[0]
            conv_e = unit_converter(energy[i], 'Ha', target='ASE')[0]
            conv_f = unit_converter(forces[i], 'Ha/Bohr', target='ASE')[0]
            conv_s = unit_converter(stress[i], 'Ha/Bohr^3', target='ASE')[0]

            atoms = create_ASE_object(atomic_numbers,
                                      positions=conv_p,
                                      cell=conv_c,
                                      energy=conv_e,
                                      forces=conv_f,
                                      stresses=conv_s)
            atoms.set_array('spinat', spinat)
            atoms_list.append(atoms)

        self.atoms_list = atoms_list
        return atoms_list



"""
// Copyright (C) 2022-2024 MLACS group (AC)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

import numpy as np
from ase.neighborlist import NeighborList
from ase.units import kB


# ========================================================================== #
def compute_pdf(configurations,
                temperature=None,
                weights=None,
                fname=None,
                **kwargs):
    '''
    Function to compute the pdf

    Parameters
    ----------

    configurations: list of ase atoms object

    Optional
    --------
    temperature: float
       The temperature of the simulation, needed for the force sampling method
    weights: list of float
       The weights of the configurations, should sum up to one
    rcut: float
       The cutoff for the distances, default to 5.0 angstrom
    method: str
       Method to compute the pdf.
       Available options are 'count' and 'force_sampling'
    nbins: int
       Number of distance bins
    '''
    # Initialize some variables and inputs
    rcut = kwargs.get('rcut', 5.0)
    method = kwargs.get('method', 'force_sampling')
    nbins = kwargs.get('nbins', 250)

    nconfs = len(configurations)
    natoms = len(configurations[0])
    volume = configurations[0].get_volume()

    # To avoid complication, we add normalized weights if not given
    if weights is None:
        weights = np.ones(nconfs) / nconfs

    if method == 'force_sampling':
        if temperature is None:
            msg = 'The temperature is needed to compute the pdf ' + \
                  'with the force sampling method'
            raise ValueError(msg)

    elements = np.array(sorted(set(configurations[0].get_chemical_symbols())))
    nelements = len(elements)
    nint = int(nelements * (nelements - 1) / 2) + nelements

    # Create a matrix to place the interactions in the right place
    R, C = np.triu_indices(nelements)
    idx_int = np.zeros((nelements, nelements), dtype=int)
    idx_int[R, C] = np.arange(nint)
    idx_int[C, R] = np.arange(nint)

    # Initialize the histogram
    count, edges = np.histogram([1.0], bins=nbins, range=(0.0, rcut))
    count = np.zeros((nint, count.shape[0]))

    points = edges[:-1] - (edges[:-1] - edges[1:]) * 0.5

    # Prepare the neighbor list
    nl = NeighborList([rcut / 2.] * natoms, self_interaction=False)
    # Loop on configurations
    for iconf in range(nconfs):
        if np.abs(weights[iconf]) >= 1e-8:
            atoms = configurations[iconf]
            nl.update(atoms)

            # We need the chemical symbols to be able to compute a
            # per-interaction g(R)
            symbols = np.array(atoms.get_chemical_symbols())

            if method == 'force_sampling':
                forces = atoms.get_forces()

            # Loop on atoms
            for iat in range(natoms):
                jat_idx, offsets = nl.get_neighbors(iat)

                # Compute the distances between atom i and its neighbor
                rij = atoms.positions[jat_idx] + \
                    np.dot(offsets, atoms.cell) - atoms.positions[iat]
                dij = np.sqrt(np.sum(rij**2, axis=1))

                iat_el = np.where(elements == symbols[iat])[0][0]

                count_iat = np.zeros_like(count)
                # If counting method, we use the histogram function of numpy
                if method == 'count':
                    for el in range(nelements):
                        el_idx = np.where(symbols[jat_idx] == elements[el])[0]
                        count_iat_tmp, edges_iat = np.histogram(
                                                      dij[el_idx],
                                                      bins=nbins,
                                                      range=(0.0, rcut))
                        inte = idx_int[iat_el, el]
                        count_iat[inte] = count_iat_tmp

                if method == 'force_sampling':
                    f_jat = forces[jat_idx]
                    tmp = -0.5 * (forces[iat] - f_jat) * \
                        rij / (dij**3).repeat(3).reshape(-1, 3)
                    # Loop on the pdf distances
                    for n in range(len(points)):
                        for el in range(nelements):
                            # We need the interactions between the elements
                            # of iat and element el
                            tmp_el_idx = np.where(
                                symbols[jat_idx] == elements[el])[0]
                            tmp_el = tmp[tmp_el_idx]
                            # inte is the index allowing to place
                            # the interaction the right place
                            inte = idx_int[iat_el, el]
                            # We extract only the distances for
                            # this type of interaction
                            dij_el = dij[tmp_el_idx]
                            # In the force-sampling formulation,
                            # all atoms within a distance inferior
                            # to R contributes
                            count_iat[inte, n] = np.sum(
                                tmp_el[dij_el < points[n]])

                # Finally we weight the contribution
                # The 2 is to account for the bothways=False in the neigh-list
                count += 2 * count_iat * weights[iconf]

    # For the multielement case, we compute the total g(R) here
    if nelements > 1:
        count = np.vstack((count, np.sum(count, axis=0)))

    # Compute the normalization factor
    norm = volume / (4 * np.pi)
    for i in range(nint):
        idx = np.where(i == idx_int)[0]
        if len(idx) > 1:
            el1 = np.array(elements[idx[0]])
            nat1 = len(np.where(el1 == atoms.get_chemical_symbols())[0])
            el2 = np.array(elements[idx[1]])
            nat2 = len(np.where(el2 == atoms.get_chemical_symbols())[0])
        else:
            el1 = np.array(elements[idx[0]])
            nat1 = len(np.where(el1 == atoms.get_chemical_symbols())[0])
            nat2 = nat1

        if method == 'count':
            count[i] *= norm / (points**2 * nat1 * nat2)
        if method == 'force_sampling':
            count[i] *= norm / (kB * temperature * nat1 * nat2)

    if nelements > 1:
        if method == 'count':
            count[-1] *= norm / (points**2 * natoms**2)
        if method == 'force_sampling':
            count[-1] *= norm / (kB * temperature * natoms**2)

    """
    # Normalize the pdf
    if method == 'count':
        count  *= norm / (points**2) * natoms
    elif method == 'force_sampling':
        count  *= norm / (kB * temperature)

    # For the multielement case, we compute the total g(R) here
    if nelements > 1:
        count = np.vstack((count, np.sum(count,axis=0)))
    """

    pdf = np.vstack((points, count)).T

    # If asked, we write the pdf
    if fname is not None:
        header = "Distances            "
        for i in range(nint):
            idx = np.where(i == idx_int)[0]
            if len(idx) > 1:
                header += f"{elements[idx[0]]}-{elements[idx[1]]}             "
            else:
                header += f"{elements[idx[0]]}-{elements[idx[0]]}             "
        if nelements > 1:
            header += "tot"
        np.savetxt(fname, pdf, header=header, fmt="%20.16f")
    return pdf



"""
// Copyright (C) 2022-2024 MLACS group (CD)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

from pathlib import Path

from ..utilities.plots import HistPlot


def main(args, parser):
    ncname = args.file
    path = Path().absolute()
    ncpath = str(path / ncname)
    ncplot = HistPlot(ncpath=ncpath)
    boolean_no_show = args.noshow
    boolean_show = not boolean_no_show
    if args.iteration is not None:
        args.iteration = [int(x) for x in args.iteration]

    ncplot.plot_neff(show=boolean_show, savename=args.save,
                     selected_iter=args.iteration)


class CLICommand:
    """
    Plot evolution of number of effective configurations, from HIST file.

    Examples:

        $ mlacs plot_neff *HIST.nc
        $ mlacs plot_neff *HIST.nc  -i 3 4 5     [Plot distribution weights
                                                  of MLACS iterations i=3,4,5]
    """
    @staticmethod
    def add_arguments(parser):
        parser.add_argument('file', help="Full name of *HIST.nc file")
        parser.add_argument("-s", "--save", default='',
                            help="Name of the file to save the plot. "
                                 "Default None")
        parser.add_argument("--noshow", action="store_true",
                            help="To disable the visualisation of the plot")
        parser.add_argument('-i', '--iteration', nargs='+', default=None,
                            help="List of MLACS iteration index ")

    @staticmethod
    def run(args, parser):
        main(args, parser)



"""
// Copyright (C) 2022-2024 MLACS group (PR)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

import numpy as np
try:  # Scipy >= 1.6.0
    from scipy.integrate import cumulative_trapezoid as cumtrapz
except ImportError:  # Scipy < 1.6.0
    from scipy.integrate import cumtrapz
from ase.units import GPa

from ..core.manager import Manager
from ..utilities.io_lammps import LammpsBlockInput

from .thermostate import ThermoState
from .solids import EinsteinSolidState
from .liquids import UFLiquidState
from .thermoint import ThermodynamicIntegration


# ========================================================================== #
# ========================================================================== #
class PressureScalingState(ThermoState):
    """
    Class for performing thermodynamic integration for a
    range of pressure using pressure scaling (in NPT).

    Parameters
    ----------
    atoms: :class:`ase.Atoms`
        ASE atoms object on which the simulation will be performed

    pair_style: :class:`str`
        pair_style for the LAMMPS input

    pair_coeff: :class:`str` or :class:`list` of :class:`str`
        pair_coeff for the LAMMPS input

    fcorr1: :class:`float` or ``None``
        First order cumulant correction to the free energy, in eV/at,
        to be added to the results.
        If ``None``, no value is added. Default ``None``.

    fcorr2: :class:`float` or ``None``
        Second order cumulant correction to the free energy, in eV/at,
        to be added to the results.
        If ``None``, no value is added. Default ``None``.

    p_start: :class:`float` (optional)
        Initial pressure of the simulation, in GPa. Default ``0``.

    p_end: :class:`float` (optional)
        Final pressure of the simulation, in GPa. Default ``10``.

    g_init: :class:`float` (optional)
        Free energy of the initial temperature, in eV/at. Default ``None``.

    phase: :class:`str`
        The phase of the system for which the free energy is computed.
        This input is used to compute the reference free energy at the
        starting pressure using a EinsteinSolidState object for solids
        and a UFLiquidState object for liquids. Can be either 'solid'
        or 'liquid'

    ninstance: :class:`int` (optional)
        If a reference Free energy calculation has to be done before
        the temperature sweep
        Settles the number of forward and backward runs. Default ``1``.

    dt: :class:`int` (optional)
        Timestep for the simulations, in fs. Default ``1.5``

    damp : :class:`float` (optional)
        Damping parameter. If ``None``, a damping parameter of a
        hundred time the timestep is used.

    temperature: :class:`float` or ``None``
        Temperature of the simulation.
        Default ``300``.

    pdamp : :class:`float` (optional)
        Damping parameter for the barostat. Default 1000 times ``dt`` is used.
        Default ``None``.

    nsteps: :class:`int` (optional)
        Number of production steps. Default ``10000``.

    nsteps_eq: :class:`int` (optional)
        Number of equilibration steps. Default ``5000``.

    nsteps_averaging: :class:`int` (optional)
        Number of steps done to equilibrate the system at the start pressure.
        Default `10000`

    gjf: :class:`bool`
        Whether to use the GJF integrator, if the Langevin thermostat is used.
        Default `True`

    rng: :class:`RNG object`
        Rng object to be used with the Langevin thermostat.
        Default correspond to :class:`numpy.random.default_rng()`

    langevin: :class:`bool`
        Whether to use a langevin thermostat. Default `True`

    logfile : :class:`str` (optional)
        Name of the file for logging the MLMD trajectory.
        If ``None``, no log file is created. Default ``None``.

    trajfile : :class:`str` (optional)
        Name of the file for saving the MLMD trajectory.
        If ``None``, no traj file is created. Default ``None``.

    interval : :class:`int` (optional)
        Number of steps between log and traj writing. Override
        loginterval and trajinterval. Default ``50``.

    loginterval : :class:`int` (optional)
        Number of steps between MLMD logging. Default ``50``.
    """
    def __init__(self,
                 atoms,
                 pair_style,
                 pair_coeff,
                 fcorr1=None,
                 fcorr2=None,
                 p_start=0,
                 p_end=10,
                 g_init=None,
                 phase=None,
                 ninstance=1,
                 dt=1,
                 damp=None,
                 temperature=300,
                 pdamp=None,
                 nsteps=10000,
                 nsteps_eq=5000,
                 nsteps_averaging=10000,
                 gjf=True,
                 rng=None,
                 langevin=True,
                 logfile=True,
                 trajfile=True,
                 interval=500,
                 loginterval=50,
                 **kwargs):

        fname = f"PressureScaling_P{p_start}_P{p_end}GPa_T{temperature}K"
        kwargs.setdefault('folder', fname)

        super().__init__(atoms,
                         pair_style,
                         pair_coeff,
                         dt=dt,
                         nsteps=nsteps,
                         nsteps_eq=nsteps_eq,
                         nsteps_averaging=nsteps_averaging,
                         rng=rng,
                         langevin=langevin,
                         logfile=logfile,
                         trajfile=trajfile,
                         interval=interval,
                         **kwargs)

        self.atoms = atoms
        self.fcorr1 = fcorr1
        self.fcorr2 = fcorr2
        self.p_start = p_start
        self.p_end = p_end
        self.g_init = g_init
        self.phase = phase
        self.ninstance = ninstance
        self.damp = damp
        self.temperature = temperature
        self.pdamp = pdamp
        self.nsteps = nsteps
        self.nsteps_eq = nsteps_eq
        self.gjf = gjf
        self.dt = dt
        self.rng = rng
        if self.rng is None:
            self.rng = np.random.default_rng()

# ========================================================================== #
    @Manager.exec_from_path
    def run(self):
        """
        """
        if self.g_init is None:
            self.run_single_ti()

        self.run_dynamics(self.atoms, self.pair_style, self.pair_coeff)

# ========================================================================== #
    def run_single_ti(self):
        """
        Free energy calculation before sweep
        """
        folder = "Fe_ref"
        if self.phase == 'solid':
            self.state = EinsteinSolidState(self.atoms,
                                            self.pair_style,
                                            self.pair_coeff,
                                            self.temperature,
                                            self.p_start,
                                            self.fcorr1,
                                            self.fcorr2,
                                            k=None,
                                            dt=self.dt,
                                            folder=folder)
        elif self.phase == 'liquid':
            self.state = UFLiquidState(self.atoms,
                                       self.pair_style,
                                       self.pair_coeff,
                                       self.temperature,
                                       self.p_start,
                                       self.fcorr1,
                                       self.fcorr2,
                                       dt=self.dt,
                                       folder=folder)

        self.ti = ThermodynamicIntegration(self.state,
                                           ninstance=self.ninstance,
                                           logfile='FreeEnergy.log')
        self.g_init = self.ti.run().mean(axis=1)[0]
        return self.g_init

# ========================================================================== #
    def _get_block_thermostat(self, eq):
        """
        """
        if self.damp is None:
            self.damp = "$(100*dt)"

        if self.pdamp is None:
            self.pdamp = "$(1000*dt)"

        temp = self.temperature
        self.info_dynamics["temperature"] = temp
        langevinseed = self.rng.integers(1, 9999999)

        block = LammpsBlockInput("thermostat", "Integrators")
        block("timestep", f"timestep {self.dt / 1000}")
        block("momenta", "velocity all create " +
              f"{temp} {langevinseed} dist gaussian")
        return block

# ========================================================================== #
    def _get_block_traj(self, el):
        """
        """
        if self.trajfile:
            block = LammpsBlockInput("dump", "Dumping")
            txt = f"dump dum1 all custom {self.loginterval} {self.trajfile} "
            txt += "id type xu yu zu vx vy vz fx fy fz "
            txt += "element"
            block("dump", txt)
            block("dump_modify1", "dump_modify dum1 append yes")
            txt = "dump_modify dum1 element " + " ".join([p for p in el])
            block("dump_modify2", txt)
            return block
        else:
            pass

# ========================================================================== #
    def _get_neti(self):
        """
        """
        li = 1
        lf = self.p_end
        temp = self.temperature

        blocks = []

        block0 = LammpsBlockInput("eq fwd", "Equilibration before fwd rs")
        line = f"fix f1 all npt temp {temp} {temp} {self.damp} "
        line += f"iso {self.p_start*10000} {self.p_start*10000} {self.pdamp}"
        block0("eq fwd npt", line)
        block0("run eq fwd", f"run {self.nsteps_eq}")
        block0("unfix eq fwd", "unfix f1")
        blocks.append(block0)

        block1 = LammpsBlockInput("fwd", "Forward Integration")
        block1("lambda fwd", "variable lambda equal " +
               f"ramp({li*10000},{lf*10000})")
        block1("pp", "variable pp equal " +
               f"ramp({self.p_start*10000},{self.p_end*10000})")
        block1("fwd npt", f"fix f2 all npt temp {temp} {temp} {self.damp} " +
               f"iso {self.p_start*10000} {self.p_end*10000} {self.pdamp}")
        block1("write fwd", "fix f3 all print 1 " +
               "\"$(pe/atoms) ${pp} ${vol} ${lambda}\" screen no " +
               "append forward.dat title \"# de                  " +
               " pressure  vol         lambda\"\n")
        block1("run fwd", f"run {self.nsteps}")
        block1("unfix fwd npt", "unfix f2")
        block1("unfix f3", "unfix f3 ")
        blocks.append(block1)

        block2 = LammpsBlockInput("eq bwd", "Equilibration before bwd rs")
        block2("eqbwd npt", f"fix f1 all npt temp {temp} {temp} {self.damp} " +
               f"iso {self.p_end*10000} {self.p_end*10000} {self.pdamp}")
        block2("run eq bwd", f"run {self.nsteps_eq}")
        block2("unfix eq bwd", "unfix f1")
        blocks.append(block2)

        block3 = LammpsBlockInput("bwd", "Backward Integration")
        block3("lambda bwd", "variable lambda equal " +
               f"ramp({lf*10000},{li*10000})")
        block3("pp", "variable pp equal " +
               f"ramp({self.p_end*10000},{self.p_start*10000})")
        block3("bwd npt", f"fix f2 all npt temp {temp} {temp} {self.damp} " +
               f"iso {self.p_end*10000} {self.p_start*10000} {self.pdamp}")
        block3("write bwd", "fix f3 all print 1 " +
               "\"$(pe/atoms) ${pp} ${vol} ${lambda}\" screen no " +
               "append backward.dat title \"# de                 " +
               " pressure  vol         lambda\"\n")
        blocks.append(block3)
        return blocks

# ========================================================================== #
    @Manager.exec_from_path
    def postprocess(self):
        """
        Compute the free energy from the simulation
        """
        natoms = len(self.atoms)

        # Get data
        _, fp, fvol, _ = np.loadtxt("forward.dat", unpack=True)
        _, bp, bvol, _ = np.loadtxt("backward.dat", unpack=True)

        # pressure contribution
        fvol = fvol / natoms
        bvol = bvol / natoms

        fp = fp / 10000 * GPa
        bp = bp / 10000 * GPa

        # Integrate the forward and backward data
        wf = cumtrapz(fvol, fp, initial=0)
        wb = cumtrapz(bvol[::-1], bp[::-1], initial=0)
        # Compute the total work
        work = (wf + wb) / 2

        pressure = np.linspace(self.p_start, self.p_end, len(work))

        free_energy = self.g_init + work

        results = np.array([pressure, free_energy]).T
        header = "p [GPa]    G [eV/at]"
        fmt = "%10.6f    %10.6f"
        np.savetxt("free_energy.dat", results, header=header, fmt=fmt)

# ========================================================================== #
    def log_recap_state(self):
        """
        """
        msg = "Thermodynamic Integration using Pressure Scaling\n"
        msg += f"Starting pressure :          {self.p_start} GPa\n"
        msg += f"Stopping pressure :          {self.p_end} GPa\n"
        msg += f"Pressure damping :              {self.pdamp} fs\n"
        msg += f"Temperature damping :              {self.damp} fs\n"
        msg += f"Timestep :                      {self.dt} fs\n"
        msg += f"Number of steps :               {self.nsteps}\n"
        msg += f"Number of equilibration steps : {self.nsteps_eq}\n"
        return msg



"""
// Copyright (C) 2022-2024 MLACS group (AC)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

import pytest
import numpy as np

from ase.io import read

from ... import context  # noqa
from mlacs.core import PathAtoms
from mlacs.utilities import interpolate_points as intpts


def issame(a, b):
    return np.all(np.abs(a - b) < 1e-8)


def test_path_atoms_img(root):
    """
    """
    f = root / 'reference_files'
    at = read(f / 'Database_NEB_path.xyz', index=':')
    neb = PathAtoms([at[0], at[1]])
    assert len(neb.images) == 2
    assert neb.nreplica == 2
    neb.images = at
    assert len(neb.images) == 8
    assert neb.nreplica == 8
    assert issame(neb.imgxi, np.linspace(0, 1, 8))
    assert len(neb.imgE) == 8
    assert neb.imgC.shape == (8, 3, 3)
    assert neb.imgR.shape == (8, 107, 3)
    assert neb.imgE[0] == neb.imgE[-1]
    assert neb.masses[0] == 1.0


def test_path_atoms_xi(root):
    """
    """
    f = root / 'reference_files'
    at = read(f / 'Database_NEB_path.xyz', index=':')

    obj = [PathAtoms(at), PathAtoms(at, xi=0.5), PathAtoms(at, mode=0.5),
           PathAtoms(at, xi=0.8), PathAtoms(at, mode='rdm')]
    obj[-2].xi = 0.5
    obj[-1].xi = 0.5

    for o in obj:
        assert o.xi == 0.5
        assert o.splined == obj[0].splined
        assert o.splE == obj[0].splE
        assert issame(o.splC, obj[0].splC)
        assert issame(o.splR, obj[0].splR)
        assert issame(o.splDR, obj[0].splDR)
        assert issame(o.splD2R, obj[0].splD2R)

    for o in obj:
        assert o.xi == o.splined.info['reaction_coordinate']


@pytest.mark.slow
def test_path_atoms_xi_array(root):
    """
    """
    f = root / 'reference_files'
    at = read(f / 'Database_NEB_path.xyz', index=':')

    xi = np.linspace(0, 1, 5)
    obj = [PathAtoms(at, xi=xi), PathAtoms(at, xi=xi.tolist()),
           PathAtoms(at, mode=xi), PathAtoms(at, mode=xi.tolist()),
           PathAtoms(at, xi=0.8), PathAtoms(at, mode='rdm')]
    obj[-2].xi = xi
    obj[-1].xi = xi

    for o in obj:
        assert issame(o.xi, xi)
        assert len(o.splined) == len(obj[0].splined)
        assert issame(o.splE, obj[0].splE)
        assert issame(o.splC, obj[0].splC)
        assert issame(o.splR, obj[0].splR)
        assert issame(o.splDR, obj[0].splDR)
        assert issame(o.splD2R, obj[0].splD2R)

    for o in obj:
        at_xi = np.r_[[a.info['reaction_coordinate'] for a in o.splined]]
        assert issame(o.xi, at_xi)


def test_path_atoms_mode(root):
    """
    """
    f = root / 'reference_files'
    at = read(f / 'Database_NEB_path.xyz', index=':')

    mode = ['saddle', 'rdm', 'rdm_memory', 'rdm_true', 0.8,
            np.linspace(0, 1, 20), [0.0, 0.4, 0.8, 1.0]]
    for m in mode:
        neb = PathAtoms(at, mode=m)
        spl = neb.splined

        # Test on the only atom moving.
        ref = np.r_[[intpts(neb.imgxi, neb.imgR[:, 0, i],
                     neb.xi, 0) for i in range(3)]]
        assert issame(ref, neb.splR[0])
        if not isinstance(m, (list, np.ndarray)):
            assert issame(ref, spl.positions[0])



"""
// Copyright (C) 2022-2024 MLACS group (AC)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

import pytest

from ase.build import bulk
from ase.io import read
from ase.calculators.emt import EMT

from ... import context  # noqa
from mlacs.mlip import SnapDescriptor, LinearPotential
from mlacs.state import NebLammpsState, PafiLammpsState
from mlacs import OtfMlacs


@pytest.fixture
def expected_folder(expected_folder_base):
    return expected_folder_base


@pytest.fixture
def expected_files():
    files = ["MLACS.log", "Training_configurations.traj",
             "MLIP-Energy_comparison.dat", "MLIP-Forces_comparison.dat",
             "MLIP-Stress_comparison.dat", "Trajectory_1.traj",
             "Trajectory_2.traj", "Trajectory_3.traj",
             "Trajectory_1_potential.dat", "Trajectory_2_potential.dat",
             "Trajectory_3_potential.dat"]
    return files


@pytest.mark.slow
@pytest.mark.skipif(context.has_lammps_nompi(),
                    reason="Lammps needs mpi to run PAFI")
def test_mlacs_pafi_multi(root, treelink):
    atoms = bulk("Ag", cubic=True).repeat(3)
    nebat = [atoms.copy(), atoms.copy()]
    nebat[0].pop(0)
    nebat[1].pop(1)
    # Check that the first atom is the one we started with
    assert len(nebat[0]) == len(nebat[1])
    natoms = len(nebat[-1])
    nstep = 3
    calc = EMT()

    mlip_params = dict(twojmax=4)
    desc = SnapDescriptor(atoms, 4.2, mlip_params)
    mlip = LinearPotential(desc, folder="Snap")

    atoms = []
    state = []
    nimages = 6
    xi = [0.01, 0.6, 1.1]
    temp = [100, 200, 300]
    for x, t in zip(xi, temp):
        neb = NebLammpsState(nebat, nimages=nimages, xi=x)
        state.append(PafiLammpsState(t, mep=neb, nsteps_eq=2, nsteps=100))
        atoms.append(nebat[0])

    sampling = OtfMlacs(atoms, state, calc, mlip, neq=5)
    sampling.run(nstep)

    for folder in treelink["folder"]:
        assert (root / folder).exists()

    for file in treelink["files"]:
        assert (root / file).exists()

    for i in range(1, 4):
        traj = read(root / f"Trajectory_{i}.traj", ":")
        assert len(traj) == nstep
        # Check that the first atom is the one we started with
        assert traj[0] == atoms[i-1]
        # Check that the system didn't change in the process
        for at in traj:
            assert len(at) == natoms



"""
// Copyright (C) 2022-2024 MLACS group (PR, AC)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

from .thermoint import ThermodynamicIntegration
from .solids import EinsteinSolidState
from .liquids import UFLiquidState
from .reversible_scaling import ReversibleScalingState
from .pressure_scaling import PressureScalingState

__all__ = ["ThermodynamicIntegration",
           "EinsteinSolidState",
           "UFLiquidState",
           "ReversibleScalingState",
           "PressureScalingState"]



"""
// Copyright (C) 2022-2024 MLACS group (AC)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

import numpy as np
from ase.calculators.singlepoint import SinglePointCalculator
from ase.units import kB, fs

hbar = 6.582119514e-16 * 1e15 * fs  # from eV.s to eV.(ASE time units)


# ========================================================================== #
def compute_centroid_atoms(confs, temperature):
    """
    Function to compute the centroid

    Parameters
    ----------

    confs: :class:`list` of :class:`ase.Atoms`
        The configurations of the quantum polymer
    temperature: :class:`float`

    Returns
    -------

    atoms: :class:`ase.Atoms`
        The centroid of the quantum polymer
    """
    nbead = len(confs)
    atoms = confs[0].copy()
    natoms = len(atoms)
    masses = confs[0].get_masses()
    momenta = np.zeros((natoms, 3))

    kBT = kB * temperature

    pos = np.zeros((nbead, natoms, 3))
    forces = np.zeros((nbead, natoms, 3))
    epot = np.zeros(nbead)
    stress = np.zeros((nbead, 6))
    cell = np.zeros((nbead, 3, 3))
    for ibead, at in enumerate(confs):
        pos[ibead] = at.get_positions()
        forces[ibead] = at.get_forces()
        epot[ibead] = at.get_potential_energy()
        stress[ibead] = at.get_stress()
        cell[ibead] = at.get_cell()

    cpos = pos.mean(axis=0)
    cforces = forces.mean(axis=0)
    cekin = 1.5 * natoms * kBT - 0.5 * np.sum((pos - cpos) * forces) / nbead
    cenergy = epot.mean()  # + cekin
    cstress = stress.mean(axis=0)
    ccell = cell.mean(axis=0)
    momenta[0, 0] = np.sqrt(2*cekin*masses[0])

    atoms.set_positions(cpos)
    atoms.set_cell(ccell, True)
    atoms.set_momenta(momenta)

    calc = SinglePointCalculator(atoms,
                                 energy=cenergy,
                                 forces=cforces,
                                 stress=cstress)
    atoms.calc = calc
    return atoms



"""
// Copyright (C) 2022-2024 MLACS group (AC)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""
import pytest

from pathlib import Path
from pytest_console_scripts import ScriptRunner


def test_mlacs_cli(script_runner: ScriptRunner) -> None:
    result = script_runner.run(['mlacs', '--version'])
    assert result.returncode == 0
    assert result.stdout == 'mlacs-1.0.1\n'
    assert result.stderr == ''
    script_runner.run('mlacs --version', shell=True, check=True)

    result = script_runner.run(['mlacs', '--help'])
    assert result.returncode == 0
    assert 'MLACS command line tool' in result.stdout.split('\n')
    assert result.stderr == ''
    functions = ['correlation', 'plot_error', 'plot_weights',
                 'plot_thermo', 'plot_neff']
    for f in functions:
        assert f in result.stdout
    script_runner.run('mlacs --help', shell=True, check=True)


@pytest.mark.parametrize("option", ['correlation', 'plot_error'])
def test_mlacs_cli_comparison(option, script_runner: ScriptRunner) -> None:
    refdir = Path().absolute() / 'reference_files' / 'Plot'
    result = script_runner.run(['mlacs', option, '--noshow',
                                f'{refdir}/MLIP-Forces_comparison.dat'])
    assert result.returncode == 0
    assert result.stderr == ''


def test_mlacs_cli_weights(script_runner: ScriptRunner) -> None:
    refdir = Path().absolute() / 'reference_files' / 'Plot'
    result = script_runner.run(['mlacs', 'plot_neff', '--noshow',
                                f'{refdir}/test_CLI_WEIGHTS.nc'])
    assert result.returncode == 0
    assert result.stderr == ''
    assert (Path().absolute() / 'plot_neff.pdf').exists()
    (Path().absolute() / 'plot_neff.pdf').unlink()


@pytest.mark.parametrize("option", ['plot_thermo', 'plot_weights'])
def test_mlacs_cli_hist(option, script_runner: ScriptRunner) -> None:
    refdir = Path().absolute() / 'reference_files' / 'Plot'
    result = script_runner.run(['mlacs', option, '--noshow',
                                f'{refdir}/test_CLI_HIST.nc'])
    assert result.returncode == 0
    assert result.stderr == ''
    assert (Path().absolute() / f'{option}.pdf').exists()
    (Path().absolute() / f'{option}.pdf').unlink()



"""
// Copyright (C) 2022-2024 MLACS group (PR, GA, AC)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""
import copy


import numpy as np

from ..core import Manager
from ..utilities import save_cwd
from ..utilities.log import ThermoLog
from .thermostate import ThermoState


# ========================================================================== #
# ========================================================================== #
class ThermodynamicIntegration(Manager):
    """
    Class to handle a series of thermodynamic integration on sampled states

    Parameters
    ----------
    thermostate: :class:`thermostate` or :class:`list` of :class:`thermostate`
        State(s) for which the thermodynamic integration should be performed
    ninstance: : class:`int`
        Numer of forward and backward to be performed per state, default 1
    logfile: :class:`str` (optional)
        Name of the logfile. Default ``\"ThermoInt.log\"``
    workdir: :class:`str`(optional)
        Name of the root folder in which the simulations will be performed.
        Default `ThermoInt`
    """
    def __init__(self,
                 thermostate,
                 ninstance=1,
                 logfile="ThermoInt.log",
                 workdir='ThermoInt',
                 **kwargs):

        Manager.__init__(self, workdir=workdir, **kwargs)

        self.workdir.mkdir(exist_ok=True, parents=True)

        self.log = ThermoLog(str(self.workdir / logfile))
        self.ninstance = ninstance
        self.logfile = logfile

        # Create list of thermostate
        if isinstance(thermostate, ThermoState):
            thermostate.workdir = self.workdir
            self.state = [thermostate]
        elif isinstance(thermostate, list):
            self.state = thermostate
            for st in self.state:
                st.workdir = self.workdir
        else:
            msg = "state should be a ThermoState object or " + \
                  "a list of ThermoState objects"
            raise TypeError(msg)
        self.nstate = len(self.state)
        self.recap_state()

# ========================================================================== #
    @Manager.exec_from_path
    def run(self):
        """
        Launch the simulation

        Returns
        -------

        fe: :class:`np.ndarray`
            The free energy of each instance of each state, in an array of
            shape (nstate, ninstance). In eV/at
        """
        values = np.zeros((self.nstate, self.ninstance))
        with save_cwd():
            for istate in range(self.nstate):
                for i in range(self.ninstance):
                    worker = copy.deepcopy(self.state[istate])
                    worker.folder += f"/for_back{i+1}"
                    worker.rng = np.random.default_rng()
                    worker.run()
                    msg, fe = worker.postprocess()
                    values[istate, i] = fe
                    self.log.logger_log.info(msg)

        for istate in range(self.nstate):
            femean = values[istate].mean()
            ferr = values[istate].std()
            msg = f"Free Energy mean and error for state {istate+1}:\n"
            msg += f"- Mean: {femean:10.6f}\n"
            msg += f"- Error: {ferr:10.6f}\n"
            self.log.logger_log.info(msg)

        self.log.write_footer()
        return values

# ========================================================================== #
    def recap_state(self):
        """
        """
        msg = f"Total number of state : {self.nstate}, "
        msg += f"each state will be run {self.ninstance} times\n"
        for istate in range(self.nstate):
            msg += f"State {istate+1}/{self.nstate} :\n"
            msg += self.state[istate].log_recap_state()
            msg += "\n\n"
        self.log.logger_log.info(msg)



"""
// Copyright (C) 2022-2024 MLACS group (AC, RB, ON)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

from pathlib import Path
import numpy as np
from ase.atoms import Atoms
from ..core.manager import Manager

from .weighting_policy import WeightingPolicy


# ========================================================================== #
# ========================================================================== #
class FixedWeight(WeightingPolicy):
    """
    Class that gives a static weight to the first few configurations and then
    use the given WeightingPolicy for the other configurations.
    Can be used to give a fixed weight to the training configurations

    Parameters
    ----------
    subweight: :class:`WeightingPolicy`
        Weight of the configurations after
        Default UniformWeight

    static_weight: :class:`np.array`
        Weights of the first configuration. The number of configuration is
        given by the length of the array. The weight must be normalized
        Default [0]
    """
    def __init__(self, static_weight=np.array([0]), subweight=None,
                 energy_coefficient=1.0, forces_coefficient=1.0,
                 stress_coefficient=1.0):

        assert np.sum(static_weight) <= 1.0
        self.subweight = subweight
        self.static_weight = static_weight
        self.nstatic = len(static_weight)
        self.remaining = 1-np.sum(static_weight)
        self.matsize = []
        self.weight = np.array([])

        if subweight is None:
            self.subweight = UniformWeight(
                energy_coefficient=energy_coefficient,
                forces_coefficient=forces_coefficient,
                stress_coefficient=stress_coefficient)
        self.energy_coefficient = self.subweight.energy_coefficient
        self.forces_coefficient = self.subweight.forces_coefficient
        self.stress_coefficient = self.subweight.stress_coefficient

# ========================================================================== #
    @Manager.exec_from_subsubdir
    def compute_weight(self, coef=None, predict=None, **kwargs):
        """
        Compute Uniform Weight taking into account nthrow :
        """
        self.subweight.workdir = self.workdir
        self.subweight.folder = self.folder
        self.subweight.subfolder = self.subfolder

        subweight_name = type(self.subweight).__name__
        header2 = ""

        if len(self.matsize) == (self.nstatic+1):  # Exactly 1 conf not static
            self.weight = np.append(self.static_weight, self.remaining)
        elif len(self.matsize) > self.nstatic:
            tmp, fn = self.subweight.compute_weight(coef=coef,
                                                    predict=predict,
                                                    **kwargs)
            header2 += tmp
            dynamic_w = self.remaining * self.subweight.weight
            self.weight = np.append(self.static_weight, dynamic_w)
        else:
            curr_weight = self.static_weight[:len(self.matsize)]
            if np.sum(curr_weight) == 0:  # A niche bug
                curr_weight = np.ones(len(self.matsize))
            self.weight = curr_weight/np.sum(curr_weight)

        header = f"Using Fixed weighting and {subweight_name}\n"
        header += f"{header2}\n"
        if Path("MLIP.weight").exists():
            Path("MLIP.weight").unlink()
        np.savetxt("MLIP.weight", self.weight, header=header, fmt="%25.20f")
        return header, "MLIP.weight"

# ========================================================================== #
    def update_database(self, atoms):
        """
        Update the database.
        """
        if isinstance(atoms, Atoms):
            atoms = [atoms]

        if len(self.matsize) > self.nstatic:
            self.subweight.update_database(atoms)
        elif len(self.matsize) + len(atoms) > self.nstatic:
            idx = len(self.matsize) + len(atoms) - self.nstatic
            self.subweight.update_database(atoms[-idx:])
        else:  # Static weights
            pass

        self.matsize.extend([len(a) for a in atoms])


# ========================================================================== #
# ========================================================================== #
class EnergyBasedWeight(WeightingPolicy):
    """
    Class that gives weight according to w_n = C/[E_n - E_min + delta]**2
    where C is a normalization constant.

    Parameters
    ----------
    delta: :class:`float`
        Shift to avoid overweighting of the ground state (eV/at)
        Default 1.0
    """
    def __init__(self, delta=1.0, energy_coefficient=1.0,
                 forces_coefficient=1.0, stress_coefficient=1.0,
                 database=None, weight=None):
        self.delta = delta
        self.energies = []
        WeightingPolicy.__init__(
                self,
                energy_coefficient=energy_coefficient,
                forces_coefficient=forces_coefficient,
                stress_coefficient=stress_coefficient,
                database=database, weight=weight)

# ========================================================================== #
    @Manager.exec_from_subsubdir
    def compute_weight(self, coef=None, predict=None, **kwargs):
        """
        Compute Uniform Weight taking into account nthrow :
        """
        if Path("MLIP.weight").exists():
            Path("MLIP.weight").unlink()
        emin = min(self.energies)
        w = np.array([1/(en - emin + self.delta)**2 for en in self.energies])
        self.weight = w / np.sum(w)

        header = "Using EnergyBased weighting\n"
        np.savetxt("MLIP.weight", self.weight, header=header, fmt="%25.20f")
        return header, "MLIP.weight"

# ========================================================================== #
    def update_database(self, atoms):
        """
        Update the database.
        """
        if isinstance(atoms, Atoms):
            atoms = [atoms]
        self.energies.extend([a.get_potential_energy()/len(a) for a in atoms])
        self.matsize.extend([len(a) for a in atoms])


# ========================================================================== #
# ========================================================================== #
class UniformWeight(WeightingPolicy):
    """
    Class that gives uniform weight in MLACS.

    Parameters
    ----------

    nthrow: :class:`int`
        Number of configurations to ignore when doing the fit.
        Three cases:

        1. If nconf > 2*nthrow, remove the nthrow first configuration
        2. If nthrow < nconf < 2*nthrow, remove the nconf-nthrow first conf
        3. If nconf < nthrow, keep all conf
    """

    def __init__(self, nthrow=0, energy_coefficient=1.0,
                 forces_coefficient=1.0, stress_coefficient=1.0,
                 database=None, weight=None):
        self.nthrow = nthrow
        WeightingPolicy.__init__(
                self,
                energy_coefficient=energy_coefficient,
                forces_coefficient=forces_coefficient,
                stress_coefficient=stress_coefficient,
                database=database, weight=weight)

# ========================================================================== #
    @Manager.exec_from_subsubdir
    def compute_weight(self, coef=None, predict=None, **kwargs):
        """
        Compute Uniform Weight taking into account nthrow :
        """
        if Path("MLIP.weight").exists():
            Path("MLIP.weight").unlink()

        nconf = len(self.matsize)
        to_remove = 0
        if nconf > 2*self.nthrow:
            to_remove = self.nthrow
        elif nconf > self.nthrow:
            to_remove = nconf-self.nthrow

        w = np.ones(nconf-to_remove) / (nconf-to_remove)
        w = np.r_[np.zeros(to_remove), w]
        self.weight = w

        header = "Using Uniform weighting\n"
        np.savetxt("MLIP.weight", self.weight, header=header, fmt="%25.20f")
        return header, "MLIP.weight"

# ========================================================================== #
    def update_database(self, atoms):
        """
        Update the database.
        """
        if isinstance(atoms, Atoms):
            atoms = [atoms]
        self.matsize.extend([len(a) for a in atoms])


# ========================================================================== #
# ========================================================================== #
class IncreasingWeight(WeightingPolicy):
    """
    Class that gives increasing weight with the index of a configuration
    in MLACS.
    This weighting policy has been though for structural optimization.

    Parameters
    ----------

    nthrow: :class:`int`
        Number of configurations to ignore when doing the fit.
        Three cases :

         1. If nconf > 2*nthrow, remove the nthrow first configuration
         2. If nthrow < nconf < 2*nthrow, remove the nconf-nthrow first conf
         3. If nconf < nthrow, keep all conf
    """

    def __init__(self, nthrow=0, power=1, energy_coefficient=1.0,
                 forces_coefficient=1.0, stress_coefficient=1.0,
                 database=None, weight=None):
        self.nthrow = nthrow
        self.power = power
        WeightingPolicy.__init__(
                self,
                energy_coefficient=energy_coefficient,
                forces_coefficient=forces_coefficient,
                stress_coefficient=stress_coefficient,
                database=database, weight=weight)

# ========================================================================== #
    @Manager.exec_from_subsubdir
    def compute_weight(self, coef, f_mlipE, **kwargs):
        """
        Compute Increasing Weight taking into account nthrow :
        """
        fname = "MLIP.weight"
        if (filepath := Path(fname)).exists():
            filepath.unlink()

        nconf = len(self.matsize)
        to_remove = 0
        if nconf > 2*self.nthrow:
            to_remove = self.nthrow
        elif nconf > self.nthrow:
            to_remove = nconf-self.nthrow

        w = (np.arange(nconf-to_remove, dtype=float) + 1)**self.power
        w /= w.sum()
        w = np.r_[np.zeros(to_remove), w]
        self.weight = w

        header = "Using Increasing weighting\n"
        np.savetxt(fname, self.weight, header=header, fmt="%25.20f")
        return header, fname

# ========================================================================== #
    def update_database(self, atoms):
        """
        Update the database.
        """
        if isinstance(atoms, Atoms):
            atoms = [atoms]
        self.matsize.extend([len(a) for a in atoms])



"""
// Copyright (C) 2022-2024 MLACS group (AC)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

from pathlib import Path
import shutil

import numpy as np
from ase.build import bulk
from ase.calculators.singlepoint import SinglePointCalculator
from ase.calculators.emt import EMT
from ase.calculators.lammpsrun import LAMMPS

from ... import context  # noqa
from mlacs.mlip import SnapDescriptor
from mlacs.mlip import LinearPotential


def create_dum_data(atoms):
    at = atoms.copy()
    at.rattle(0.01, seed=np.random.randint(1, 99999))
    energy = np.random.random()
    forces = np.random.random((len(atoms), 3))
    stress = np.random.random(6)
    calc = SinglePointCalculator(atoms, energy=energy, forces=forces,
                                 stress=stress)
    calc = EMT()
    at.calc = calc
    return at


def test_linear_potential():
    """
    TODO find other things to test
    Here, we just test if everything runs alright
    """
    root = Path()
    folder = 'Snap'
    snapfold = root / folder
    if snapfold.exists():
        shutil.rmtree(snapfold)

    at = bulk("Cu").repeat(2)

    params = dict(twojmax=4)
    desc = SnapDescriptor(at, 3.0, params)

    # First we test if OLS functions
    print(snapfold.absolute())
    mlip = LinearPotential(
        desc, workdir=root, folder=folder, subfolder='LeastSquare')
    fakeat = []
    for _ in range(5):
        fakeat.append(create_dum_data(at))

    mlip.update_matrices(fakeat)
    mlip.train_mlip()

    shutil.rmtree(snapfold)

    # And with ridge regression
    mlip_params = dict(method="ridge")
    mlip = LinearPotential(desc, parameters=mlip_params,
                           workdir=root, folder=folder, subfolder='Ridge')

    mlip.update_matrices(fakeat)
    mlip.train_mlip()

    # Let's check that what we compute with the matrix is also what
    # we get with LAMMPS

    calc = LAMMPS(pair_style=mlip.pair_style,
                  pair_coeff=mlip.pair_coeff)
    mlip_fakeat = []
    emat = []
    fmat = []
    smat = []
    for at in fakeat:
        at0 = at.copy()
        at0.calc = calc
        at0.get_potential_energy()
        mlip_fakeat.append(at0)

        etmp, ftmp, stmp = mlip.predict(at)
        emat.append(etmp)
        fmat.append(ftmp)
        smat.append(stmp)

    emat = np.array(emat)
    fmat = np.array(fmat)
    smat = np.array(smat)

    elammps = np.array([at.get_potential_energy() for at in mlip_fakeat])
    flammps = np.array([at.get_forces() for at in mlip_fakeat])
    flammps = np.reshape(flammps, np.shape(fmat))
    slammps = np.array([at.get_stress() for at in mlip_fakeat])

    # TODO check if it's a writing problem or if we have a slight difference in
    # How we compute stuff wrt LAMMPS
    assert np.allclose(elammps, emat, atol=1e-0)
    assert np.allclose(flammps, fmat, atol=1e-0)
    assert np.allclose(slammps, smat, atol=1e-0)

    shutil.rmtree(snapfold)
    for file in ["Energy", "Forces", "Stress"]:
        (root / f"MLIP-{file}_comparison.dat").unlink()



import re
import pytest

import numpy as np
from ase.build import bulk
from ase.io import read
from ase.calculators.emt import EMT
from ase.optimize import BFGS, BFGSLineSearch
from ase.filters import UnitCellFilter
from ase.units import GPa

from ... import context  # noqa
from mlacs.state import OptimizeAseState
from mlacs.mlip import SnapDescriptor, LinearPotential, DeltaLearningPotential
from mlacs import MlMinimizer


@pytest.fixture
def files_with_prefix():
    files = []
    fltr = [None, UnitCellFilter, UnitCellFilter]
    press = [None, 0.0, 5.0]
    algo = [None, BFGS, BFGSLineSearch]
    for f, p, a in zip(fltr, press, algo):
        if a is None:
            name = "None"
        else:
            name = a.__name__
        if f is None:
            fname = "None"
        else:
            fname = f.__name__
        files.append(f'{name}_{p}_{fname}.traj')
        files.append(f'{name}_{p}_{fname}_potential.dat')
    return files


@pytest.fixture
def expected_folder():
    folder = ["MolecularDynamics", "Snap"]
    return folder


@pytest.fixture
def expected_files(files_with_prefix):
    files = ["MLACS.log", "MLACS.log0001", "MLACS.log0002",
             "Training_configurations.traj",
             "MLIP-Energy_comparison.dat", "MLIP-Forces_comparison.dat",
             "MLIP-Stress_comparison.dat"]
    files.extend(files_with_prefix)
    return files


def test_mlacs_optimize(root, treelink):

    atoms = bulk("Cu", cubic=True).repeat(2)
    atoms.pop(0)
    natoms = len(atoms)
    nstep = 20
    calc = EMT()

    mlip_params = dict(twojmax=4)
    desc = SnapDescriptor(atoms, 4.2, mlip_params)
    mlip = LinearPotential(desc, folder="Snap")
    ps, cs = 'zbl 1.0 2.0', ['* * 29 29']
    dmlip = DeltaLearningPotential(mlip, pair_style=ps, pair_coeff=cs)

    etol = 0.01
    ftol = 0.01
    stol = 0.01

    prefix = []
    fltr = [None, UnitCellFilter, UnitCellFilter]
    press = [None, 0.0, 5.0]
    algo = [None, BFGS, BFGSLineSearch]
    for f, p, a in zip(fltr, press, algo):
        if a is None:
            name = "None"
        else:
            name = a.__name__
        if f is None:
            fname = "None"
        else:
            fname = f.__name__
        prefix.append(f'{name}_{p}_{fname}')
        if p is not None:
            press = p * GPa
        state = OptimizeAseState(optimizer=a, filters=f,
                                 fltr_parameters=dict(scalar_pressure=press,
                                                      cell_factor=10),
                                 prefix=prefix[-1])
        sampling = MlMinimizer(atoms, state, calc, dmlip, etol, ftol, stol)
        sampling.run(nstep)

    for folder in treelink["folder"]:
        assert (root / folder).exists()

    for file in treelink["files"]:
        assert (root / file).exists()

    for p in prefix:
        traj = read(root / f"{p}.traj", ":")
        # Check that the first atom is the one we started with
        assert traj[0] == atoms
        # Check that the system didn't change in the process
        for at in traj:
            assert len(at) == natoms
        # Check that the criterion on forces is achieved
        assert np.max(traj[-1].get_forces()) <= ftol
        # Check that volume is constant
        if 'None' in p:
            assert traj[-1].get_volume() == atoms.get_volume()
        # Check that the pressure is consistent with the target
        else:
            pres = -traj[-1].get_stress()[:3].mean() / GPa
            ptarg = float(re.findall("[0-9]", p)[0])
            assert np.isclose(pres, ptarg, atol=1e-1)



"""
// Copyright (C) 2022-2024 MLACS group (AC)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

import pytest

from ase.build import bulk
from ase.calculators.emt import EMT

from ... import context  # noqa
from mlacs.mlip import TensorpotPotential, AceDescriptor
from mlacs.mlip import EnergyBasedWeight
from mlacs.state import LammpsState
from mlacs import OtfMlacs


"""
Example of a MLACS simulation using ACE
This test is slow.
Maybe put it outside the testsuite
"""


@pytest.fixture
def expected_folder(expected_folder_base):
    return ["MLIP", "MolecularDynamics"]


@pytest.fixture
def expected_files(expected_files_base):
    return expected_files_base


@pytest.mark.slow
@pytest.mark.skipif(context.has_pyace(), reason="Need Python-Ace to use ACE")
def test_ace_restart(root, treelink):
    import pandas as pd

    # Parameters --------------------------------------------------------------
    temperature = 300  # K
    nconfs = 2
    nsteps = 10
    nsteps_eq = 10
    neq = 1
    cell_size = 2
    dt = 1.5  # fs
    rcut = 6

    # Supercell creation ------------------------------------------------------
    atoms = bulk('Cu', cubic=True).repeat(cell_size)
    calc = EMT()
    state = LammpsState(temperature, dt=dt, nsteps=nsteps,
                        nsteps_eq=nsteps_eq)

    weighter = EnergyBasedWeight(delta=0.025, stress_coefficient=0)
    fitting_dict = {'maxiter': 200, 'fit_cycles': 1, 'repulsion': 'auto',
                    'optimizer': 'BFGS',
                    'optimizer_options': {'disp': True, 'gtol': 0, 'xrtol': 0}}
    ace_descriptor = AceDescriptor(atoms=atoms,
                                   free_at_e={'Cu': 0},
                                   tol_e=10.0,  # meV/at
                                   tol_f=500.0,  # meV/ang
                                   nworkers=4,
                                   fitting_dict=fitting_dict,
                                   rcut=rcut)
    mlip = TensorpotPotential(descriptor=ace_descriptor,
                              weight=weighter)
    sampling = OtfMlacs(atoms, state, calc, mlip, neq=neq)
    sampling.run(nconfs)

    # Running Restart Lammps State
    state = LammpsState(temperature, dt=dt, nsteps=nsteps,
                        nsteps_eq=nsteps_eq)
    weighter = EnergyBasedWeight(delta=0.025, stress_coefficient=0)
    ace_descriptor = AceDescriptor(atoms=atoms,
                                   free_at_e={'Cu': 0},
                                   tol_e=10.0,  # meV/at
                                   tol_f=500.0,  # meV/ang
                                   nworkers=4,
                                   fitting_dict=fitting_dict,
                                   rcut=rcut)
    mlip = TensorpotPotential(descriptor=ace_descriptor,
                              weight=weighter)
    sampling = OtfMlacs(atoms, state, calc, mlip, neq=neq)
    sampling.run(nconfs)
    df = pd.read_pickle("MLIP/ACE.pckl.gzip", compression="gzip")
    # traj = read("Trajectory.traj", index=":")
    assert sum(sampling.nconfs) == 4
    assert len(df) == 4
    # assert len(traj) == 4  # BUG : To fix



"""
// Copyright (C) 2022-2024 MLACS group (AC)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

import pytest

import numpy as np
from ase.build import bulk
from ase.io import read
from ase.calculators.emt import EMT

from ... import context  # noqa
from mlacs.mlip import SnapDescriptor, LinearPotential
from mlacs.state import LammpsState
from mlacs import OtfMlacs


@pytest.fixture
def expected_folder(expected_folder_base):
    return expected_folder_base


@pytest.fixture
def expected_files():
    files = ["MLACS.log", "Training_configurations.traj",
             "MLIP-Energy_comparison.dat", "MLIP-Forces_comparison.dat",
             "MLIP-Stress_comparison.dat", "Trajectory_1.traj",
             "Trajectory_2.traj", "Trajectory_3.traj",
             "Trajectory_1_potential.dat", "Trajectory_2_potential.dat",
             "Trajectory_3_potential.dat"]
    return files


def test_mlacs_multistate(root, treelink):

    atoms = bulk("Cu", cubic=True).repeat(2)
    natoms = len(atoms)
    nstep = 4
    nconfs = 3
    nconfs_init = 1
    calc = EMT()

    mlip_params = dict(twojmax=4)
    desc = SnapDescriptor(atoms, 4.2, mlip_params)
    mlip = LinearPotential(desc, workdir=root, folder="Snap")

    atoms = []
    state = []
    acell = [3.300, 3.200, 3.100]
    temp = [100, 200, 300]
    press = [-1, 0, 1]
    for t, p, a in zip(temp, press, acell):
        state.append(LammpsState(t, p, nsteps_eq=2, nsteps=100))
        atoms.append(bulk("Cu", cubic=True, a=a).repeat(2))
    nstate = len(state)

    sampling = OtfMlacs(atoms, state, calc, mlip, neq=5, workdir=root)
    sampling.run(nstep)

    for folder in treelink["folder"]:
        assert (root / folder).exists()

    for file in treelink["files"]:
        assert (root / file).exists()

    for i in range(1, 4):
        traj = read(root / f"Trajectory_{i}.traj", ":")
        assert len(traj) == nstep
        # Check that the first atom is the one we started with
        assert traj[0] == atoms[i-1]
        # Check that the system didn't change in the process
        for at in traj:
            assert len(at) == natoms

    ml_energy = np.loadtxt(root / "MLIP-Energy_comparison.dat")
    ml_forces = np.loadtxt(root / "MLIP-Forces_comparison.dat")
    ml_stress = np.loadtxt(root / "MLIP-Stress_comparison.dat")

    nconfs = nstate * (nconfs + nconfs_init)
    assert ml_energy.shape == (nconfs, 2)
    assert ml_forces.shape == (nconfs * natoms * 3, 2)
    assert ml_stress.shape == (nconfs * 6, 2)



"""
// Copyright (C) 2022-2024 MLACS group (AC, RB, ON)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

import os
from concurrent.futures import ThreadPoolExecutor

import numpy as np

from ase.atoms import Atoms
from ase.io import read, Trajectory
from ase.io.formats import UnknownFileTypeError
from ase.calculators.calculator import Calculator
from ase.calculators.singlepoint import SinglePointCalculator

from .core import Manager
from .mlip import LinearPotential, MliapDescriptor
from .calc import CalcManager
from .state import StateManager
from .utilities.log import MlacsLog
from .utilities import create_random_structures, save_cwd
from .utilities.io_abinit import OtfMlacsHist
from .properties import PropertyManager, RoutinePropertyManager


# ========================================================================== #
# ========================================================================== #
class Mlas(Manager):
    r"""
    A Learn on-the-fly simulation constructed in order to sample approximate
    distribution

    Parameters
    ----------

    atoms: :class:`ase.Atoms` or :class:`list` of :class:`ase.Atoms`
        the atom object on which the simulation is run.

    state: :class:`StateManager` or :class:`list` of :class:`StateManager`
        Object determining the state to be sampled

    calc: :class:`ase.calculators` or :class:`CalcManager`
        Class controlling the potential energy of the system
        to be approximated.
        If a :class:`ase.calculators` is attached, the :class:`CalcManager`
        is automatically created.

    mlip: :class:`MlipManager` (optional)
        Object managing the MLIP to approximate the real distribution
        Default is a LammpsMlip object with a snap descriptor,
        ``5.0`` angstrom rcut with ``8`` twojmax.

    prop: :class:`PropertyManager` or :class:`list` or :class:`CalcProperty`
    (optional)
        Object managing the MLIP to approximate the real distribution
        Default is a LammpsMlip object with a snap descriptor,
        ``5.0`` angstrom rcut with ``8`` twojmax.

    neq: :class:`int` (optional)
        The number of equilibration iteration. Default ``10``.

    confs_init: :class:`int` or :class:`list` of :class:`ase.Atoms` (optional)
        If :class:`int`, Number of configurations used to train a preliminary
        MLIP. The configurations are created by rattling the first structure.
        If :class:`list` of :class:`ase.Atoms`, The atoms that are to be
        computed in order to create the initial training configurations.
        Default ``None``.

    std_init: :class:`float` (optional)
        Variance (in :math:`\mathring{a}^2`) of the displacement
        when creating initial configurations.
        Default :math:`0.05 \mathring{a}^2`

    keep_tmp_mlip: :class:`Bool` (optional)
        Keep every generated MLIP. If True and using MBAR, a restart will
        recalculate every previous MLIP.weight using the old coefficients.
        Default ``True``.

    workdir: :class:`str` (optional)
        The directory in which to run the calculation.

    prefix: :class:`str` (optional)
        The prefix to prepend the name of the States files.

    ncprefix: :class:`str` (optional)
        The prefix to prepend the name of the *HIST.nc file.
        Script name format: ncprefix + scriptname + '_HIST.nc'.
        Default `''`.

    ncformat: :class:`str` (optional)
        The format of the *HIST.nc file. One of the five flavors of netCDF
        files format available in netCDF4 python package: 'NETCDF3_CLASSIC',
        'NETCDF3_64BIT_OFFSET', 'NETCDF3_64BIT_DATA','NETCDF4_CLASSIC',
        'NETCDF4'.
        Default ``NETCDF3_CLASSIC``.
    """

    def __init__(self,
                 atoms,
                 state,
                 calc,
                 mlip=None,
                 prop=None,
                 neq=10,
                 confs_init=None,
                 std_init=0.05,
                 keep_tmp_mlip=True,
                 workdir='',
                 prefix='',
                 ncprefix='',
                 ncformat='NETCDF3_CLASSIC'):

        Manager.__init__(self, workdir=workdir, prefix=prefix)

        # Initialize working directory
        self.workdir.mkdir(exist_ok=True, parents=True)
        self.ncfile = None

        ##############
        # Check inputs
        ##############
        self.keep_tmp_mlip = keep_tmp_mlip
        self._initialize_state(state, atoms, neq, prefix)
        self._initialize_calc(calc)
        self._initialize_mlip(mlip)
        # self._initialize_properties(prop)

        # Miscellanous initialization
        self.rng = np.random.default_rng()

        # Check if trajectory files already exists
        self.launched = self._check_if_launched()

        # Create Abinit-style *HIST.nc file of netcdf format
        self.ncfile = OtfMlacsHist(ncprefix=ncprefix,
                                   workdir=workdir,
                                   ncformat=ncformat,
                                   launched=self.launched,
                                   atoms=self.atoms)
        if self.ncfile.unique_atoms_type:
            self._initialize_routine_properties()

        self._initialize_logger()

        # Initialize momenta and parameters for initial/training configs
        if not self.launched:
            self._initialize_momenta()
            self.traj = []
            self.confs_init = confs_init or []
            self.std_init = std_init
            self.nconfs = [0] * self.nstate
            self.uniq_at = None
            self.idx_computed = None
            self.nb_distinct_conf = None

        # Reinitialize everything from the trajectories
        # Compute fitting data - get trajectories - get current configurations
        else:
            self.restart_from_traj()

        self.step = 0
        self.log._delimiter()
        self._write("Starting the simulation", True)

# ========================================================================== #
    @Manager.exec_from_workdir
    def run(self, nsteps=100):
        """
        Run the algorithm for nsteps
        """
        isearlystop = False
        while self.step < nsteps:
            if self._check_early_stop():
                isearlystop = True
                break
            self.log.init_new_step(self.step)
            if not self.launched:
                self._run_initial_step()
                self.step += 1
            else:
                step_done = self._run_step()
                if not step_done:
                    pass
                else:
                    self.step += 1

        self.log.write_end(isearlystop)
        # Here we have to add some info about the end of the simulation
        self.log.write_footer()

# ========================================================================== #
    def _run_step(self):
        """
        Run a regular step of the algorithm.

        A regular step consists in:
           - Train (fit) the MLIP,
           - Run MLMD with trained MLIP during nsteps,
           - Compute the true potential energy/forces/stresses,
           - Update database and save new configs.

        Returns
        ----------

        `partial_success`: :class:`bool`
            `True` if at least one state has been successfully computed by the
            true potential after MLMD, `False` otherwise.
            See also _handle_potential_error() method.
        """

        eq = self._is_eq_step()

        # Train MLIP
        self._write("Training new MLIP")
        self._set_mlip_subfolder()
        self.mlip.train_mlip()

        # Run MLMD with as many threads as there are states
        self._write("Running MLMD")
        atoms_mlip = self._create_mlip_atoms(eq)
        futures = []
        with save_cwd(), ThreadPoolExecutor() as executor:
            for istate in range(self.nstate):
                exe = executor.submit(self.state[istate].run_dynamics,
                                      *(atoms_mlip[istate],
                                        self.mlip.pair_style,
                                        self.mlip.pair_coeff,
                                        self.mlip.model_post,
                                        self.mlip.atom_style,
                                        eq[istate],
                                        self.mlip.get_elements()))
                futures.append(exe)
                msg = f"State {istate+1}/{self.nstate} has been launched"
                self._write(msg)
            for istate, exe in enumerate(futures):
                atoms_mlip[istate] = exe.result()
                if self.keep_tmp_mlip:
                    mm = self.mlip.descriptor.subsubdir
                    atoms_mlip[istate].info['parent_mlip'] = str(mm)
            executor.shutdown(wait=True)

        # Computing energy with true potential
        self._write("Computing energy with the True potential")
        # TODO GA: Might be better to do the threading at this level,
        #          up from calc.compute_true_potential.
        subfolder_l = [s.subfolder for s in self.state]
        step_l = [max(self.nconfs)] * self.nstate
        atoms_true = self.calc.compute_true_potential(atoms_mlip,
                                                      subfolder_l,
                                                      step=step_l)
        partial_success = self._handle_potential_error(atoms_true)

        # SinglePointCalculator to bypass the calc attach to atoms thing of ase
        for at in atoms_mlip:
            at.calc = self.mlip.get_calculator()
            at.calc = SinglePointCalculator(at.copy(),
                                            energy=at.get_potential_energy(),
                                            forces=at.get_forces(),
                                            stress=at.get_stress())

        # Update database / Save new configs in trajectory files
        attrue = self.add_traj_descriptors(atoms_true)
        for i, (attrue, atmlip) in enumerate(zip(atoms_true, atoms_mlip)):
            if attrue is not None:
                self.mlip.update_matrices(attrue)
                self.traj[i].write(attrue)
                self.atoms[i] = attrue

                prefix = self.state[i].prefix
                filepath = self.workdir / (prefix + "_potential.dat")
                with open(filepath, "a") as f:
                    f.write("{:20.15f}   {:20.15f}\n".format(
                             attrue.get_potential_energy(),
                             atmlip.get_potential_energy()))
                self.nconfs[i] += 1

        # TODO: CD: Implement groups in netcdf for Atoms with difft chem. form.
        if self.ncfile.unique_atoms_type:
            self._compute_properties()
            self._compute_routine_properties()
        self._execute_post_step()

        return partial_success

# ========================================================================== #
    def _run_initial_step(self):
        """
        Run the initial step.

        Notes
        -----

        This routine unfolds in three stages.

            (i) Compute initial configurations.
            `Initial` refers to the atomic configurations `atoms` given
            as a parameter to the `Mlas` object. These confs are assigned to
            self.atoms and self.uniq_at, and get stored in `Trajectory.traj` at
            the end of the routine.

                NB: `Trajectory.traj` becomes `Trajectory_i.traj` if there are
                several states, or more generally `prefix.traj` as defined in
                _initialize_state().

            (ii) [Conditionnal] Compute training configurations.
            `Training` refers to configurations that serve as complementary
            configurations ensuring the database is large enough to fit an
            MLIP. These confs are assigned to self.confs_init and are saved in
            `Training_configurations.traj`.

                NB: This stage is only executed if the MLIP is empty. This is
                intentional, as MLACS enables a non-empty MLIP to be loaded at
                the start of the calculation (independently of restarts). Note
                how, in that case, the `initial` configs are still computed.

                NB: There are several options to set these training confs. One
                of them is to load them from an existing *.traj file.

            (iii) Update MLIP object with new confs from step (i) and (ii).
        """
        self._write("\nRunning initial step\n")

        # Compute initial configurations
        self._get_unique_atoms()
        self._compute_initial_confs()
        self._distribute_atoms()
        self._create_initial_traj()

        # Load/compute training configurations (if empty MLIP)
        if self.mlip.nconfs == 0:
            conf_fname = str(self.workdir / "Training_configurations.traj")
            if self._is_traj_file(conf_fname):
                self._load_training_confs(conf_fname)
            else:
                self._initialize_training_confs()
                self._compute_training_confs()
                self._create_training_traj()

        # Add initial/training configurations to database
        for at in self.confs_init + self.uniq_at:
            self.mlip.update_matrices(at)

        self.launched = True

# ========================================================================== #
    def _initialize_calc(self, calc):
        """Create calculator object"""
        if isinstance(calc, Calculator):
            self.calc = CalcManager(calc)
        elif isinstance(calc, CalcManager):
            self.calc = calc
        else:
            msg = "calc should be a ase Calculator object or " + \
                  "a CalcManager object"
            raise TypeError(msg)
        self.calc.workdir = self.workdir

# ========================================================================== #
    def _initialize_mlip(self, mlip):
        """Create mlip object"""
        if mlip is None:
            descriptor = MliapDescriptor(self.atoms[0], 5.0)
            self.mlip = LinearPotential(descriptor)
        else:
            self.mlip = mlip

        self.mlip.workdir = self.workdir
        if not self.mlip.folder:
            self.mlip.folder = 'MLIP'
        self.mlip.descriptor.workdir = self.workdir
        self.mlip.descriptor.folder = self.mlip.folder
        self.mlip.weight.workdir = self.workdir
        self.mlip.weight.folder = self.mlip.folder
        self.mlip.subdir.mkdir(exist_ok=True, parents=True)

# ========================================================================== #
    def _initialize_properties(self, prop):
        """Create property object"""
        self.prop = PropertyManager(prop)

# ========================================================================== #
    def _initialize_routine_properties(self):
        """Create routine property object"""
        self.routine_prop = RoutinePropertyManager(self.ncfile,
                                                   self.launched)

# ========================================================================== #
    def _initialize_momenta(self):
        """Create property object"""
        for i in range(self._nmax):
            self.state[i].initialize_momenta(self.atoms[i])
            prefix = self.state[i].prefix
            pot_fname = self.workdir / (prefix + "_potential.dat")
            with open(pot_fname, "w") as f:
                f.write("# True epot [eV]          MLIP epot [eV]\n")
        self.prefix = ''

# ========================================================================== #
    def _initialize_state(self, state, atoms, neq, prefix='Trajectory'):
        """
        Function to initialize the state
        """
        # Put the state(s) as a list
        if isinstance(state, StateManager):
            self.state = [state]
        if isinstance(state, list):
            self.state = state
        self.nstate = len(self.state)

        for s in self.state:
            s.workdir = self.workdir
            s.folder = 'MolecularDynamics'
            if not s.subfolder:
                s.subfolder = prefix
            if not s.prefix:
                s.prefix = prefix

        if self.nstate > 1:
            for i, s in enumerate(self.state):
                s.subfolder = s.subfolder + f"_{i+1}"
                s.prefix = s.prefix + f"_{i+1}"

        self.atoms = []
        # Create list of atoms
        if isinstance(atoms, Atoms):
            for _ in range(self.nstate):
                self.atoms.append(atoms.copy())
        elif isinstance(atoms, list):
            e = "You should have 1 atoms per state"
            assert len(atoms) == self.nstate, e
            self.atoms = [at.copy() for at in atoms]
        else:
            msg = "atoms should be a ASE Atoms object or " + \
                  "a list of ASE atoms objects"
            raise TypeError(msg)
        self.atoms_start = [at.copy() for at in self.atoms]

        # Create list of neq -> number of equilibration
        # mlmd runs for each state
        if isinstance(neq, int):
            self.neq = [neq] * self.nstate
        elif isinstance(neq, list):
            assert len(neq) == self.nstate
            self.neq = self.nstate
        else:
            msg = "neq should be an integer or a list of integers"
            raise TypeError(msg)

# ========================================================================== #
    def _initialize_logger(self):
        """Initialize logger."""
        self.log = MlacsLog(str(self.workdir / "MLACS.log"), self.launched)
        # XXX: self.logger: useless line?
        # self.logger = self.log.logger_log
        self._write()
        self.log._delimiter()
        self._write("Recap of the simulation parameters", True)
        self._write()
        self._write("Recap of the states", False, True)
        for i in range(self.nstate):
            self._write(f"State {i+1}/{self.nstate} :")
            self._write(repr(self.state[i]))
            self._write()
        self._write()
        self._write("Recap of the calculator", False, True)
        msg = self.calc.log_recap_state()
        self._write(msg)
        self._write()
        self._write("Recap of the MLIP", False, True)
        self._write(repr(self.mlip))
        self._write()

        # Share logger instance to mlip and calc objects.
        self.mlip.logger = self.log
        self.calc.logger = self.log

# ========================================================================== #
    def _check_if_launched(self):
        """
        Function to check simulation restarts:
         - Check if trajectory files exist and are not empty.
         - Check if the number of configuration found is at least two.

        Returns
        ----------

        `bool`
            Boolean indicating if the conditions are met.
        """
        # Build a list of all *.traj files to test
        all_traj_files = [str(self.workdir / "Training_configurations.traj")]
        for i in range(self._nmax):
            traj_fname = str(self.workdir / (self.state[i].prefix + ".traj"))
            all_traj_files.append(traj_fname)

        # Check all possible *.traj files
        _nat_init = 0
        for traj_fname in all_traj_files:
            if self._is_traj_file(traj_fname):
                _nat_init += len(read(traj_fname, index=':'))
            else:
                return False

        return _nat_init >= 2

# ========================================================================== #
    def _is_eq_step(self):
        """
        Distinguish between equilibration and production steps for the MLMD.

        Returns
        ----------

        `eq`: :class:`list` of :class:`bool`
            Element in `eq` is True if this is an equilibration step.
        """
        self._write()
        eq = []
        for istate in range(self.nstate):
            trajstep = self.nconfs[istate]
            if self.nconfs[istate] < self.neq[istate]:
                eq.append(True)
                msg = f"Equilibration step for state {istate+1}, "
            else:
                eq.append(False)
                msg = f"Production step for state {istate+1}, "
            msg += f"{trajstep} configurations for this state"
            self._write(msg)
        self._write()
        return eq

# ========================================================================== #
    def _set_mlip_subfolder(self):
        """
        Set mlip subfolder.

        Notes
        ----------

        The branching here results from two situations:

        (1) One wants to keep all MLIP data (i.e., for each step),

        (2) One is only interested in the last MLIP data, and therefore
            routinely overwritting MLIP data is fine.
        """
        if self.keep_tmp_mlip:
            self.mlip.subfolder = f"Coef{max(self.nconfs)}"
        else:
            self.mlip.subfolder = ''

# ========================================================================== #
    def _create_mlip_atoms(self, eq):
        """
        Create MLIP atoms object.

        Returns
        ----------

        `atoms_mlip` :class:`list` of :class:`ase.Atoms`
            The list of Atoms objects on which MLMD will be performed.
            Can be either
                - the initial configs (if restart or equilibration steps)
                - the last `self.atoms` configs (reinstantiated for precaution)
        """
        atoms_mlip = [None]*self.nstate
        for istate in range(self.nstate):
            if self.state[istate].isrestart or eq[istate]:
                self._write(" -> Starting from first atomic configuration")
                atoms_mlip[istate] = self.atoms_start[istate].copy()
                self.state[istate].initialize_momenta(atoms_mlip[istate])
            else:
                atoms_mlip[istate] = self.atoms[istate].copy()
        return atoms_mlip

# ========================================================================== #
    def _handle_potential_error(self, atoms_true):
        """
        True potential error handling.

        Returns
        ----------

        `bool`
            `True` if at least one state has been successfully computed by the
            true potential after MLMD, `False` otherwise.
        """
        partial_success = True
        nerror = 0
        for i, at in enumerate(atoms_true):
            if at is None:
                msg = f"For state {i+1}/{self._nmax} calculation with " + \
                       "the true potential resulted in error " + \
                       "or didn't converge"
                self._write(msg)
                nerror += 1
        if nerror == self.nstate:
            msg = "All true potential calculations failed, " + \
                  "restarting the step"
            self._write(msg)
            partial_success = False
        return partial_success

# ========================================================================== #
    def _is_traj_file(self, trajname):
        """
        Check if Trajectory file exists and is readable.

        Parameters
        ----------

        trajname: :class:`str`
            The *.traj file being tested.

        Returns
        ----------

        `bool`
            Boolean indicating if the conditions are met.
        """
        if not os.path.isfile(trajname):
            return False
        try:
            read(trajname, index=":")
            return True
        except UnknownFileTypeError:
            return False

# ========================================================================== #
    def add_traj_descriptors(self, atoms):
        """
        Add descriptors in trajectory files.
        """
        if isinstance(atoms, Atoms):
            atoms = [atoms]
        desc = self.mlip.descriptor.compute_descriptors(atoms)
        for iat, at in enumerate(atoms):
            at.info['descriptor'] = desc[iat]
        return atoms

# ========================================================================== #
    @Manager.exec_from_workdir
    def restart_from_traj(self):
        """
        Restart a calculation from previous trajectory files
        """
        train_traj, prev_traj = self.read_traj()

        #for i in range(self._nmax):
        #    self.state[i].subsubdir.mkdir(exist_ok=True, parents=True)

        # Add the Configuration without a MLIP generating them
        if train_traj is not None:
            for i, conf in enumerate(train_traj):
                self._write(f"Configuration {i+1} / {len(train_traj)}")
                self.mlip.update_matrices(conf)  # We add training conf

        # Add all the configuration of trajectories traj
        self._write("Adding previous configuration iteratively")
        # GA: TODO We dont actually need parent_list. Remove this variable.
        parent_list, mlip_coef = self.mlip.read_parent_mlip(prev_traj)

        # Directly adding initial conf to have it once even if multistate
        atoms_by_mlip = [[] for _ in range(len(parent_list))]
        no_parent_atoms = [prev_traj[0][0]]

        for istate in range(self.nstate):
            for iconf in range(1, len(prev_traj[istate])):
                if "parent_mlip" in prev_traj[istate][iconf].info:
                    pm = prev_traj[istate][iconf].info['parent_mlip']
                    idx = parent_list.index(pm)
                    atoms_by_mlip[idx].append(prev_traj[istate][iconf])
                else:
                    no_parent_atoms.append(prev_traj[istate][iconf])

        for conf in no_parent_atoms:
            self.mlip.update_matrices(conf)

        # If the last simulation was with keep_tmp_mlip=False,
        # we put the old MLIP.model and weight in a Coef folder
        can_use_weight = self.mlip.can_use_weight
        if len(no_parent_atoms) > 1 and self.keep_tmp_mlip and can_use_weight:
            self._write("Some configuration in Trajectory have no parent_mlip")
            self._write("You should rerun this simulation with DatabaseCalc")

            fm = self.mlip.subdir / "MLIP.model"
            fw = self.mlip.subdir / "MLIP.weight"

            last_coef = max(self.nconfs)-1
            self.mlip.subfolder = f"Coef{last_coef}"

            if os.path.isfile(fm):
                if not os.path.exists(self.mlip.subsubdir):
                    self.mlip.subsubdir.mkdir(exist_ok=True, parents=True)
                    os.rename(fm, self.mlip.subsubdir / "MLIP.model")
                    os.rename(fw, self.mlip.subsubdir / "MLIP.weight")

        curr_step = 0
        for i in range(len(atoms_by_mlip)):
            curr_step += 1

            # GA: Since we don't read
#           self.mlip.subsubdir = Path(atoms_by_mlip[i][0].info['parent_mlip'])
            self.mlip.next_coefs(mlip_coef[i])
            for at in atoms_by_mlip[i]:
                self.mlip.update_matrices(at)
        self.mlip.subfolder = ''

        # Update this simulation traj
        self.traj = []
        self.atoms = []

        for i in range(self._nmax):
            traj_fname = str(self.workdir / (self.state[i].prefix + ".traj"))
            self.traj.append(Trajectory(traj_fname, mode="a"))
            self.atoms.append(prev_traj[i][-1])
        del prev_traj

# ========================================================================== #
    def read_traj(self):
        """
        Read Trajectory files from previous simulations
        """
        self._write("Adding previous configurations to the training data")

        conf_fname = str(self.workdir / "Training_configurations.traj")
        if os.path.isfile(conf_fname):
            train_traj = Trajectory(conf_fname, mode="r")
            self._write(f"{len(train_traj)} training configurations\n")
        else:
            train_traj = None

        prev_traj = []
        lgth = []
        for i in range(self._nmax):
            traj_fname = str(self.workdir / (self.state[i].prefix + ".traj"))
            prev_traj.append(Trajectory(traj_fname, mode="r"))
            lgth.append(len(prev_traj[i]))
        self.nconfs = lgth
        self._write(f"{np.sum(lgth)} configuration from trajectories")
        return train_traj, prev_traj

# ========================================================================== #
    @property
    def _nmax(self):
        return self.nstate

# ========================================================================== #
    def _check_early_stop(self):
        """
        Break the self consistent procedure.
        """
        return self.prop.check_criterion

# ========================================================================== #
    def _execute_post_step(self):
        """
        Function to execute some things that might be needed for a specific
        mlas object.
        For example, computing properties
        """
        pass

# ========================================================================== #
    def _compute_properties(self):
        """
        Function to execute and converge on specific Properties.
        For example, CalcRdf, CalcTI ...
        """
        pass

# ========================================================================== #
    def _compute_routine_properties(self):
        """Compute routine properties"""
        self.routine_prop.calc_initialize(atoms=self.atoms)
        msg = self.routine_prop.run(self.step)
        self.log.logger_log.info(msg)
        self.routine_prop.save_prop(self.step)
        self.routine_prop.save_weighted_prop(self.step, self.mlip.weight)
        self.routine_prop.save_weights(self.step,
                                       self.mlip.weight,
                                       self.ncfile.ncformat)

# ========================================================================== #
    def _get_unique_atoms(self):
        """
        Compute list of unique atoms and corresponding indices.

        Notes
        -----

        The initial `atoms` list, self.atoms, may be redundant (e.g., if state-
        averaging is sought during MD). To avoid calculating the same true
        potential several times, a list of `unique` configurations (two-by-two
        distinct `Atoms` objects) is obtained in this function.
        """
        uniq_at, idx_computed = [], []
        for istate, at in enumerate(self.atoms):
            for i, unique_at in enumerate(uniq_at):
                if at == unique_at:
                    idx_computed[i].append(istate)
                    break  # Found identical conf, exit loop
            else:  # No identical conf. found, so add to list of unique confs
                uniq_at.append(at)
                idx_computed.append([istate])
        self.idx_computed = idx_computed
        self.nb_distinct_conf = len(uniq_at)
        self.uniq_at = uniq_at

        self._write(f"Number of distinct config.: {self.nb_distinct_conf}")

# ========================================================================== #
    def _distribute_atoms(self):
        """
        Update `self.atoms` with computed properties gathered in `uniq_at`, by
        distributing unique atoms properties to all atoms.
        """
        for iun, at in enumerate(self.uniq_at):
            for icop in self.idx_computed[iun]:
                # All atoms/calc are reinstantiated
                newat = at.copy()
                calc = SinglePointCalculator(newat,
                                             energy=at.get_potential_energy(),
                                             forces=at.get_forces(),
                                             stress=at.get_stress())
                newat.calc = calc
                self.atoms[icop] = newat
                # XXX: [CD] does this reinstantiation serve any purpose at all?
                # It seems that the fate of self.atoms[*] is to be written in
                # *.traj files, but self.atoms[*] is never modified itself.
                # As a matter of fact, the instruction
                # self.atoms[icop] = at
                # where unique atoms share the same atom/calc instances, pass
                # all the tests.

# ========================================================================== #
    def _create_initial_traj(self):
        """
        Create the .traj files.
        Write the initial configurations (stored in self.atoms) in .traj files.
        """
        self._write("Creating trajectories")
        for istate in range(self.nstate):
            prefix = self.state[istate].prefix
            self.traj.append(Trajectory(prefix + ".traj", mode="w"))  # Create
            self.traj[istate].write(self.atoms[istate])  # Update
            self.nconfs[istate] += 1

# ========================================================================== #
    def _compute_initial_confs(self):
        """
        Compute true potential properties of (unique) initial configurations.
        """
        subfolder_l = ["Initial"] * self.nb_distinct_conf
        istep = np.arange(self.nb_distinct_conf, dtype=int)
        self._write("Computing true potential energy [initial configs]")
        self.uniq_at = self.calc.compute_true_potential(self.uniq_at,
                                                        subfolder_l,
                                                        istep)
        if any(at is None for at in self.uniq_at):
            msg = "True potential calculation failed or didn't converge"
            raise TruePotentialError(msg)
        self._write("Computation done")

# ========================================================================== #
    def _initialize_training_confs(self):
        """Create training configurations"""
        if self.confs_init == []:
            # By default, only one training configuration is created
            self.confs_init = create_random_structures(self.uniq_at,
                                                       self.std_init,
                                                       1)
        elif isinstance(self.confs_init, (int, float)):
            # XXX: why should it be self.atoms[0] that is passed as an
            # arg here to create_random_structures, and not uniq_at?
            self.confs_init = create_random_structures(self.atoms[0],
                                                       self.std_init,
                                                       self.confs_init)

# ========================================================================== #
    def _load_training_confs(self, conf_fname):
        """Load and log training configurations."""

        self.confs_init = read(conf_fname, index=":")
        nb_confs_found = len(self.confs_init)

        self._write(f"{nb_confs_found} training configurations found")
        self._write("Adding them to the training data")

        if nb_confs_found >= 2:
            self._write("\tNo need to start new training computations\n")
        self._write()

# ========================================================================== #
    def _compute_training_confs(self):
        """
        Compute true potential properties of training configurations.
        """
        nstate = len(self.confs_init)
        subfolder_l = ["Training"] * nstate
        istep = np.arange(nstate, dtype=int)
        self._write("\nComputing true potential energy [training configs]")
        self.confs_init = self.calc.compute_true_potential(self.confs_init,
                                                           subfolder_l,
                                                           istep)
        if any(at is None for at in self.confs_init):
            msg = "True potential calculation failed or didn't converge"
            raise TruePotentialError(msg)
        self._write("Computation done")

# ========================================================================== #
    def _create_training_traj(self):
        """
        Create the training *.traj file.
        Write the initial configurations (stored in self.confs_init) in *.traj.
        """
        conf_fname = str(self.workdir / "Training_configurations.traj")
        init_traj = Trajectory(conf_fname, mode="w")
        for conf in self.confs_init:
            init_traj.write(conf)
        self._write()

# ========================================================================== #
    def _write(self, msg="", center=False, underline=False):
        self.log.write(msg, center, underline)


class TruePotentialError(Exception):
    """
    To be raised if there is a problem with the true potential
    """
    pass



"""
// Copyright (C) 2022-2024 MLACS group (CD)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

import pytest
from pathlib import Path
import os
import netCDF4 as nc
import numpy as np

from ase.build import bulk
from ase.calculators.emt import EMT
from ase.units import Bohr, Hartree

from mlacs.mlip import SnapDescriptor
from mlacs.mlip import LinearPotential
from mlacs.mlip import MbarManager
from mlacs.state import LammpsState
from mlacs.properties import CalcExecFunction
from mlacs import OtfMlacs
from mlacs.utilities.io_abinit import MlacsHist, AbinitNC


from ... import context  # noqa


@pytest.mark.skipif(context.has_netcdf(),
                    reason="You need the netCDF4 package to run the test.")
def test_basic_hist():
    """
    Check that a '*HIST_nc' file is created, with required format.
    Check that it stores some basic thermodynamic variables.
    Check that it can be read with the hist_read() method.
    """
    pytest_path = os.getenv('PYTEST_CURRENT_TEST')
    root = Path(pytest_path).parents[0].absolute()
    test_wkdir = root / 'tmp_hist_dir1'

    nb_mlacs_iter1 = 5
    at = bulk("Cu", cubic=True).repeat(2)
    calc = EMT()
    T = 400
    P = 0
    nsteps = 10
    parameters = {"twojmax": 6}
    descriptor = SnapDescriptor(at, parameters=parameters)
    mbar = MbarManager(parameters={"solver": "L-BFGS-B"})
    mlip_mbar = LinearPotential(descriptor, weight=mbar)
    target_nc_format = 'NETCDF4'
    nb_states = 2
    states = list(LammpsState(T, P, nsteps=nsteps) for i in range(nb_states))
    properties = [CalcExecFunction('get_kinetic_energy')]
    dyn1 = OtfMlacs(at,
                    states,
                    calc,
                    mlip_mbar,
                    properties,
                    neq=0,
                    workdir=test_wkdir,
                    ncprefix='obj1',
                    ncformat=target_nc_format)
    dyn1.run(nb_mlacs_iter1)

    # Check that file exists
    script_name = 'obj1_' + 'test_hist_file'
    script_name += '_HIST.nc'
    path_name = test_wkdir / script_name
    assert path_name.is_file()

    # Check that some basic properties are stored, and have proper features
    with nc.Dataset(str(path_name), 'r') as ncfile:
        assert ncfile.file_format == target_nc_format
        vol_var = ncfile['vol']
        vol_data = vol_var[:].data
        vol_unit = vol_var.unit

        assert isinstance(vol_var[:], np.ma.MaskedArray)
        assert isinstance(vol_data, np.ndarray)
        assert vol_data.shape == (nb_states*(nb_mlacs_iter1-1),)
        assert vol_unit == 'Bohr^3'
        etotal_var = ncfile['etotal']
        etotal_unit = etotal_var.unit
        assert etotal_unit == 'Ha'

    # Check that the MlacsHist class is running properly
    ncfile = MlacsHist(ncpath=path_name)
    var_names = ncfile.get_var_names()
    assert isinstance(var_names, list)
    dict_var_units = ncfile.get_units()
    assert isinstance(dict_var_units, dict)
    var_dim_dict = ncfile.var_dim_dict
    dict_name_label = {x[0]: label for label, x in var_dim_dict.items()}
    assert 'temper' in dict_name_label

    # Check that weights from HIST and from MLIP folder are the same
    weights = ncfile.read_obs('weights')
    weights_meta = ncfile.read_obs('weights_meta')
    weights_idx = weights_meta[:, 0]
    dict_weights = {}
    idx_bounds = np.argwhere(weights_idx == 1.0)[:, 0]
    for ii in range(len(idx_bounds)-1):
        iter_mlacs = ii+1
        i1, i2 = idx_bounds[ii], idx_bounds[ii+1]
        dict_weights[iter_mlacs] = [weights_idx[i1:i2], weights[i1:i2]]
    last_weights = dict_weights[iter_mlacs][1]
    N_eff = np.sum(last_weights)**2 / np.sum(last_weights**2)
    if nb_mlacs_iter1 >= 5:
        subfold = "Coef" + str(nb_mlacs_iter1-2)
        p = test_wkdir / 'MLIP' / subfold
        test_weights = np.loadtxt(p / "MLIP.weight")
        # The first two training confs are excluded from the HIST file
        test_w = test_weights[2:]
        test_w /= np.sum(test_w)
        nb_eff_test = np.sum(test_w)**2 / np.sum(test_w**2)
        assert np.round(N_eff, 5) == np.round(nb_eff_test, 5)


@pytest.mark.skipif(context.has_netcdf(),
                    reason="You need the netCDF4 package to run the test.")
def test_distinct_hist():
    """
    Check that a distinct OtfMlacs object with different features
    is saved in a separate *HIST.nc file.
    In particular, for NETCDF3_CLASSIC format, a *WEIGHTS.nc file is created.
    """
    pytest_path = os.getenv('PYTEST_CURRENT_TEST')
    root = Path(pytest_path).parents[0].absolute()
    test_wkdir2 = root / 'tmp_hist_dir2'

    nb_mlacs_iter2 = 2
    at = bulk("Cu", cubic=True).repeat(2)
    calc = EMT()
    T = 400
    P = 0
    nsteps = 10
    parameters = {"twojmax": 6}
    descriptor = SnapDescriptor(at, parameters=parameters)
    parameters = {"solver": "L-BFGS-B"}
    mlip_mbar = LinearPotential(descriptor)
    target_nc_format2 = 'NETCDF3_CLASSIC'
    nc_name2 = 'obj2_test_hist_file_HIST.nc'
    path_name2 = test_wkdir2 / nc_name2
    nc_weights = 'obj2_test_hist_file_WEIGHTS.nc'
    path_weight = test_wkdir2 / nc_weights
    nb_states2 = 2
    states2 = list(LammpsState(T, P, nsteps=nsteps) for i in range(nb_states2))
    dyn2 = OtfMlacs(at,
                    states2,
                    calc,
                    mlip_mbar,
                    neq=0,
                    workdir=test_wkdir2,
                    ncprefix='obj2',
                    ncformat=target_nc_format2)
    dyn2.run(nb_mlacs_iter2)
    assert path_name2.is_file()
    assert path_weight.is_file()
    with nc.Dataset(str(path_name2), 'r') as ncfile:
        assert ncfile.file_format == target_nc_format2
        vol_var = ncfile['vol']
        vol_data = vol_var[:].data
        assert vol_data.shape == (nb_states2*(nb_mlacs_iter2-1),)


@pytest.mark.skipif(context.has_netcdf(),
                    reason="You need the netCDF4 package to run the test.")
def test_restart_hist():
    """
    Check restart of *HIST.nc file, i.e. proper concatenation of data.
    """
    pytest_path = os.getenv('PYTEST_CURRENT_TEST')
    root = Path(pytest_path).parents[0].absolute()
    test_wkdir = root / 'tmp_hist_dir3'
    target_nc_format = 'NETCDF4'
    nb_states = 1

    nb_mlacs_iter1 = 3
    nb_mlacs_iter3 = 1
    at = bulk("Cu", cubic=True).repeat(2)
    calc = EMT()
    T = 400
    P = 0
    nsteps = 10
    parameters = {"twojmax": 6}
    descriptor = SnapDescriptor(at, parameters=parameters)
    parameters = {"solver": "L-BFGS-B"}
    mlip_mbar = LinearPotential(descriptor)
    states = list(LammpsState(T, P, nsteps=nsteps) for i in range(nb_states))
    properties = [CalcExecFunction('get_kinetic_energy')]
    dyn1 = OtfMlacs(at,
                    states,
                    calc,
                    mlip_mbar,
                    properties,
                    neq=0,
                    workdir=test_wkdir,
                    ncprefix='obj3',
                    ncformat=target_nc_format)
    dyn1.run(nb_mlacs_iter1)

    states = list(LammpsState(T, P, nsteps=nsteps) for i in range(nb_states))
    properties = [CalcExecFunction('get_kinetic_energy')]
    dyn2 = OtfMlacs(at,
                    states,
                    calc,
                    mlip_mbar,
                    properties,
                    neq=0,
                    workdir=test_wkdir,
                    ncprefix='obj3',
                    ncformat=target_nc_format)
    dyn2.run(nb_mlacs_iter3)

    script_name = 'obj3_' + 'test_hist_file'
    script_name += '_HIST.nc'
    path_name = test_wkdir / script_name
    with nc.Dataset(str(path_name), 'r') as ncfile:
        vol_var = ncfile['vol']
        vol_data = vol_var[:].data
        assert vol_data.shape == (nb_states*(nb_mlacs_iter1-1+nb_mlacs_iter3),)


@pytest.mark.skipif(context.has_netcdf(),
                    reason="You need the netCDF4 package to run the test.")
def test_hist_converter():
    """
    Check that
        (1) converter from Abinit HIST.nc to list of ase.Atoms
            [namely AbinitNC.convert_to_atoms()]
        (2) converter from list of ase.Atoms to Abinit HIST.nc
            [namely AtomsToHist]
    are inverses of each other.
    """
    # Load a reference HIST.nc file from an Abinit calculation
    workdir = str(Path() / 'reference_files/')
    prefix = 'sqa'
    suffix = 'HIST'
    hist = AbinitNC(workdir=workdir, prefix=prefix, suffix=suffix)
    hist_results = hist.read()
    hist_cell = hist_results['rprimd']
    hist_positions = hist_results['xcart']

    # Convert to ASE list of Atoms
    hist_atoms = hist.convert_to_atoms()

    # Check the conversion of units (must be ASE units)
    assert len(hist_atoms) == len(hist_positions)
    ase_energy_first_conf = hist_atoms[0].get_total_energy()
    abinit_energy_first_conf = hist_results['etotal'][0]*Hartree
    assert ase_energy_first_conf == abinit_energy_first_conf
    ase_cell_first_conf = np.array(hist_atoms[0].get_cell())
    abinit_cell_first_conf = hist_cell[0]*Bohr
    assert np.array_equal(ase_cell_first_conf, abinit_cell_first_conf)

    # Convert ASE list of Atoms back into HIST.nc format
    ncprefix = 'myconverted'
    workdir = Path('')
    obj = MlacsHist(ncprefix=ncprefix,
                    workdir=workdir,
                    atoms=hist_atoms)
    obj.convert_to_hist()

    # Check that created new HIST.nc is same as the reference one.
    hist_new = AbinitNC()
    nc_name = ncprefix + '_HIST.nc'
    nc_path = Path('') / nc_name
    hist_new.ncfile = str(nc_path)
    hist_new_results = hist_new.read()
    assert np.allclose(hist_new_results['znucl'],
                       hist_results['znucl'],
                       rtol=10**-8)
    assert np.allclose(hist_new_results['etotal'],
                       hist_results['etotal'],
                       rtol=10**-8)
    assert np.allclose(hist_new_results['xcart'],
                       hist_results['xcart'],
                       rtol=10**-8)
    assert np.allclose(hist_new_results['fcart'],
                       hist_results['fcart'],
                       rtol=10**-8)
    assert np.allclose(hist_new_results['strten'],
                       hist_results['strten'],
                       rtol=10**-8)

    nc_path.unlink()


@pytest.mark.skipif(context.has_netcdf(),
                    reason="You need the netCDF4 package to run the test.")
def test_gsr_out():
    """
    Check that reference GSR.nc/OUT.nc files can be read by AbinitNC class,
    and converted to list of ase.Atoms in ASE units.
    """

    workdir = str(Path() / 'reference_files/')
    prefix = 'abinit'

    gsr = AbinitNC(workdir=workdir, prefix=prefix, suffix='GSR')
    gsr_atoms = gsr.convert_to_atoms()

    out = AbinitNC(workdir=workdir, prefix=prefix, suffix='OUT')

    out_atoms = out.convert_to_atoms()

    assert len(gsr_atoms) == 1
    assert len(out_atoms) == 1

    gsr_cell = np.array(gsr_atoms[0].get_cell())
    gsr_results = gsr.read()
    nc_gsr_cell = gsr_results['primitive_vectors']
    assert np.allclose(gsr_cell, nc_gsr_cell*Bohr, rtol=10**-8)

    out_cell = np.array(out_atoms[0].get_cell())
    out_results = out.read()
    out_rprim = np.reshape(out_results['rprim'], (3, 3))
    nc_out_cell = out_results['acell'] * out_rprim
    assert np.allclose(out_cell, nc_out_cell*Bohr, rtol=10**-8)



"""
// Copyright (C) 2022-2024 MLACS group (AC, RB, ON)
// This file is distributed under the terms of the
// GNU General Public License, see LICENSE.md
// or http://www.gnu.org/copyleft/gpl.txt .
// For the initials of contributors, see CONTRIBUTORS.md
"""

import os
from pathlib import Path
from contextlib import contextmanager
import numpy as np

from scipy import interpolate
from scipy.integrate import simpson
from scipy.optimize import minimize

from ase.atoms import Atoms
from ase.calculators.singlepoint import SinglePointCalculator as SPC


# ========================================================================== #
def get_elements_Z_and_masses(supercell):
    '''
    Get the unique chemical symbols and atomic numbers of a supercell.
    The list are returned according to the alphabetical order of the elements.

    Parameters
    ----------
    supercell: :class:`ase.Atoms`
        ASE atoms object

    Return
    ------
    elements: :class:`list` of :class:`str`
        list of unique elements in the supercell
    Z: :class:`list` of :class:`int`
        list of unique Z in the supercell
    masses: :class:`list` of :class:`float`
        list of unique masses in the supercell
    '''
    elements = supercell.get_chemical_symbols()
    Z = supercell.get_atomic_numbers()
    masses = supercell.get_masses()
    charges = supercell.get_initial_charges()

    un_elements = sorted(set(elements))
    un_Z = []
    un_masses = []
    un_charges = []
    for iel in range(len(un_elements)):
        idx = elements.index(un_elements[iel])
        un_Z.append(Z[idx])
        un_masses.append(masses[idx])
        un_charges.append(charges[idx])

    if np.allclose(un_charges, 0.0, atol=1e-8):
        un_charges = None
    un_elements = np.array(un_elements)
    un_Z = np.array(un_Z)
    return np.array(un_elements), un_Z, un_masses, un_charges


# ========================================================================== #
def create_random_structures(atoms, std, nconfs):
    """
    Create nconfs random structures by displacing atoms around positions.

    Parameters
    ----------
    atoms: :class:`ase.Atoms` or :class:`list` of :class:`ase.Atoms`
        ASE atoms objects to be rattled
    std: :class:`float`
        Standard deviation of the gaussian used to generate
        the random displacements. In angstrom.
    nconfs: :class:`int`
        Number of configurations to generate

    Return
    ------
    confs: :class:`list` of :class:`ase.Atoms`
        Configurations with random displacements
    """
    if isinstance(atoms, Atoms):
        atoms = [atoms]
    rng = np.random.default_rng()
    confs = []
    for iat, at in enumerate(atoms):
        for i in range(nconfs):
            iatoms = at.copy()
            iatoms.rattle(stdev=std, rng=rng)
            confs.append(iatoms)
    return confs


# ========================================================================== #
def compute_correlation(data, weight=None):
    """
    Function to compute the RMSE, MAE and Rsquared

    Parameters
    ----------

    data: :class:`numpy.ndarray` of shape (ndata, 2)
        The data for which to compute the correlation.
        The first column should be the ground truth and the second column
        should be the prediction of the model
    weight: :class:`numpy.ndarray`
        Weight to be applied to compute the averages.
        Has to be a divisor of the length of the data

    Returns
    -------
        result: :class:`numpy.ndarray`
            An length 3 array with (in order) the rmse, mae and :class:`R^2`
    """
    if weight is None:  # Uniform weighting
        nconf = np.shape(data)[0]
        weight = np.ones(nconf) / nconf
    datatrue = data[:, 0]
    datatest = data[:, 1]
    assert len(datatrue) % len(weight) == 0, "Weights isn't a divisor of data"
    weight = np.repeat(weight, len(datatrue)//len(weight))

    mae = np.average(np.abs(datatrue - datatest), weights=weight)
    rmse = np.sqrt(np.average((datatrue - datatest)**2, weights=weight))
    mae = np.average(np.abs(datatrue - datatest), weights=weight)
    sse = np.average(((datatrue - datatest)**2), weights=weight)
    sst = np.average((datatrue - np.sum(datatrue*weight))**2, weights=weight)
    rsquared = 1 - sse / sst
    return np.array([rmse, mae, rsquared])


# ========================================================================== #
def create_ASE_object(atomic_numbers,
                      positions,
                      cell,
                      energy=None,
                      forces=None,
                      stresses=None):
    """
    Create ASE Atoms object.
    """
    atoms = Atoms(numbers=atomic_numbers,
                  positions=positions,
                  cell=cell,
                  pbc=True)
    atoms.calc = SPC(atoms=atoms,
                     energy=energy,
                     forces=forces,
                     stress=stresses)
    return atoms


# ========================================================================== #
def compute_averaged(traj, weights=None):
    """
    Function to compute the averaged Atoms configuration from a Trajectory.

    Parameters
    ----------

    data: :class:`ase.Trajectory`
        List of ase.Atoms object.

    Return
    ------
    confs: :class:`ase.Atoms`
        Averaged configuration
    """
    if isinstance(traj, Atoms):
        traj = [traj]
    if weights is None:
        weights = np.ones(len(traj))
    Z = traj[-1].get_atomic_numbers()
    cell = np.average([at.get_cell()[:] for at in traj],
                      axis=0, weights=weights)
    positions = np.average([at.get_positions() for at in traj],
                           axis=0, weights=weights)
    energy = np.average([at.get_potential_energy() for at in traj],
                        weights=weights)
    atoms = Atoms(numbers=Z,
                  positions=positions,
                  cell=cell,
                  pbc=True)
    calc = SPC(atoms=atoms,
               energy=energy)
    atoms.calc = calc
    return atoms.copy()


# ========================================================================== #
def interpolate_points(x, y, xf, order=0, smooth=0, periodic=0, border=None):
    """
    Interpolate points.

    Parameters
    ----------
    x : :class:`numpy.array`
        List of points to interpolate
    y : :class:`numpy.array`
        List of points to interpolate
    xf : :class:`numpy.array`
        New thiner list of points
    order : :class:`int`
        Order of the spline
    smooth : :class:`int`
        Smoothing parameter
    periodic : :class:`int`
        Activate periodic function boundary conditions
    border : :class:`bol`
        Impose a zero derivative condition at the function boundaries

    Return
    ------
    yf : :class:`list`
        List of interpolated points
    """

    def err(c, x, y, t, k):
        """The error function to minimize"""
        diff = y - interpolate.splev(x, (t, c, k))
        diff = np.einsum('...i,...i', diff, diff)
        return np.abs(diff)

    tck = interpolate.splrep(x, y, s=smooth, per=periodic)
    if border is not None:
        t, c0, k = tck
        x0 = (x[0], x[-1])
        con = {'type': 'eq',
               'fun': lambda c: interpolate.splev(x0, (t, c, k), der=1),
               }
        opt = minimize(err, c0, (x, y, t, k), constraints=con)
        copt = opt.x
        tck = (t, copt, k)

    if isinstance(xf, list):
        yf = [interpolate.splev(_, tck, der=order) for _ in xf]
        return yf
    elif isinstance(xf, np.ndarray):
        yf = [interpolate.splev(_, tck, der=order) for _ in xf]
        return yf
    else:
        return float(interpolate.splev(xf, tck, der=order))


# ========================================================================== #
def integrate_points(x, y, xf, order=0, smooth=0, periodic=0, border=None):
    """
    Interpolate points and return derivatives.

    Parameters
    ----------
    x : :class:`numpy.array`
        List of points to interpolate
    y : :class:`numpy.array`
        List of points to interpolate
    xf : :class:`numpy.array`
        New thiner list of points
    order : :class:`int`
        Order of the spline
    smooth : :class:`int`
        Smoothing parameter
    periodic : :class:`int`
        Activate periodic function boundary conditions
    border : :class:`bol`
        Impose a zero derivative condition at the function boundaries

    Return
    ------
    yf : :class:`list`
        Integral of spline from start to xf
    """

    def err(c, x, y, t, k):
        """The error function to minimize"""
        diff = y - interpolate.splev(x, (t, c, k))
        diff = np.einsum('...i,...i', diff, diff)
        return np.abs(diff)

    def integ(x, tck):
        """Integral of spline from start to x"""
        x = np.atleast_1d(x)
        prim = np.zeros(x.shape, dtype=x.dtype)
        for i in range(len(prim)):
            prim[i] = interpolate.splint(0, x[i], tck)
        return prim

    tck = interpolate.splrep(x, y, s=smooth, per=periodic)
    if border is not None:
        t, c0, k = tck
        x0 = (x[0], x[-1])
        con = {'type': 'eq',
               'fun': lambda c: interpolate.splev(x0, (t, c, k), der=1),
               }
        opt = minimize(err, c0, (x, y, t, k), constraints=con)
        copt = opt.x
        tck = (t, copt, k)
    if isinstance(xf, list):
        yf = integ(xf, tck)
        return yf
    elif isinstance(xf, np.ndarray):
        yf = integ(xf, tck)
        return yf
    else:
        return float(integ(xf, tck))


# ========================================================================== #
def normalized_integration(x, y, norm=1.0, scale=True, func=simpson):
    """
    Compute normalized integral of y to `norm`.

    Parameters
    ----------
    x : :class:`numpy.array`
    y : :class:`numpy.array`
    norm : :class:`float`
        Norm of the integral.
    scale : :class:`Bool`
        Scale x and y to the same order of magnitude to avoid numerical
        errors.
    func : :class:`scipy.integrate.func`
        Scipy function for intergration (simpson, trapz, ...).

    Return
    ------
    Scaled y
    """
    fx, fy = 1.0, 1.0
    sx, sy = x, y
    if scale:
        fx, fy = np.abs(x).max(), np.abs(y).max()
        sx, sy = x / fx, y / fy
    _norm = func(sy, sx) * fx * fy
    return y * norm / _norm


# ========================================================================== #

# GA: This context manager might be problematic when used in parallel
#     execution. We might not come back where we started!
@contextmanager
def execute_from(directory):
    """Context to work from a subfolder."""
    workdir = Path(directory)
    initial = Path.cwd().absolute()

    if not workdir.exists():
        workdir.mkdir(exist_ok=True, parents=True)

    if workdir != initial:
        os.chdir(workdir)
    try:
        yield
    finally:
        if workdir != initial:
            os.chdir(initial)


@contextmanager
def save_cwd():
    """Context to move back to the current directory after execution."""
    initial = Path.cwd().absolute()
    try:
        yield
    finally:
        cwd = Path.cwd().absolute()
        if cwd != initial:
            os.chdir(initial)


# ========================================================================== #
def create_link(fn, lk):
    """
    Creates a symbolic link lk pointing to fn
    If lk already exists, replace it
    """
    fn = Path(fn)
    lk = Path(lk)
    if os.path.isfile(lk):
        if os.path.islink(lk):  # lk is already a link
            os.remove(lk)
        else:  # lk is already a file
            return
    if not os.path.exists(fn):
        return
    src = fn.relative_to(lk.parent)
    dst = lk
    os.symlink(src, dst)


# ========================================================================== #
def read_distribution_files(filename):
    """
    Function to the distribution files of LAMMPS
    Return the averaged values and the extrem values.
    """
    yaxis, buf = None, None
    with open(filename, 'r') as r:
        for line in r:
            if line[0] == '#':
                continue
            rowdat = line.split()
            if len(rowdat) == 2:
                if yaxis is None and buf is None:
                    yaxis = None
                elif yaxis is None and buf is not None:
                    yaxis = buf
                else:
                    yaxis = np.c_[yaxis, buf]
                buf = np.zeros(int(rowdat[1]))
                xaxis = np.zeros(int(rowdat[1]))
                i = 0
                continue
            xaxis[i] = float(rowdat[1])
            buf[i] = float(rowdat[2])
            i += 1
    _gav = np.average(np.c_[yaxis, buf].T, axis=0)
    _gmin = np.min(np.c_[yaxis, buf].T, axis=0)
    _gmax = np.max(np.c_[yaxis, buf].T, axis=0)
    return xaxis, _gav, _gmin, _gmax
